#!/usr/bin/env python

from __future__ import print_function

import sys
import argparse
import re

from dark.taxonomy import (
    formatLineage, addTaxonomyDatabaseCommandLineOptions,
    parseTaxonomyDatabaseCommandLineOptions)

VERSION_REGEX = re.compile(r'\.\d+$')


def taxonomyInfo(id_, db, namesOnly, separator):
    """
    Convert the lineage information for C{id_} into a formatted string.

    @param id_: The C{str} accession number or C{int} taxonomy id to
        look up.
    @param db: The sqlite3 taxonomy database to consult.
    @param namesOnly: If C{True} only print taxonomic names.
    @param separator: The C{str} separator to put between fields. If C{None}
        return space-padded aligned columns.
    @return: A formatted C{str} for printing or C{None} if C{id_} is
        not found in the taxonomy database.
    """
    try:
        idInt = int(id_)
    except ValueError:
        lineage = db.lineage(id_)
    else:
        lineage = db.lineage(idInt)

    if lineage is None and VERSION_REGEX.search(id_) is None:
        # Try adding a version number.
        lineage = db.lineage(id_ + '.1')

    if lineage:
        return formatLineage(lineage, namesOnly, separator)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=('Print taxonomy information for accession numbers or '
                     'taxonomy ids.'))

    parser.add_argument(
        'ids', nargs='*', metavar='id',
        help=('The ids (accession numbers, names, or taxonomy ids) to print '
              'taxonomy information for. If not given, ids are read from '
              'standard input, one per line.'))

    parser.add_argument(
        '--separator',
        help=('The string to separate output columns with. Default is to '
              'print aligned output padded with multiple spaces (or a TAB if '
              '--namesOnly is used).'))

    parser.add_argument(
        '--namesOnly', default=False, action='store_true',
        help='If specified, only print the taxonomic names.')

    parser.add_argument(
        '--print', default=False, action='store_true',
        help='If specified, also print the id.')

    addTaxonomyDatabaseCommandLineOptions(parser)

    args = parser.parse_args()

    db = parseTaxonomyDatabaseCommandLineOptions(args, parser)

    if args.ids:
        ids = args.ids
    else:
        ids = (line.strip() for line in sys.stdin)

    separator = args.separator

    if args.namesOnly and separator is None:
        separator = '\t'

    for id_ in ids:
        if args.print:
            print(id_ + ':')
        result = taxonomyInfo(id_, db, args.namesOnly, separator)
        if result:
            print(result)
        else:
            print('%r was not found in the taxonomy database' % id_,
                  file=sys.stderr)



#!/usr/bin/env python

from __future__ import print_function, division

import sys
import argparse
from os.path import join
from math import log10
from subprocess import CalledProcessError
import multiprocessing
from tempfile import mkdtemp
from shutil import rmtree

from dark.dna import compareDNAReads, matchToString, AMBIGUOUS
from dark.fasta import FastaReads
from dark.reads import (Reads, addFASTACommandLineOptions,
                        parseFASTACommandLineOptions)
from dark.process import Executor
from dark.utils import parseRangeString

MAFFT_DEFAULT_ARGS = '--globalpair --maxiterate 1000 --preservecase'
MAFFT_ALGORITHMS_URL = (
    'https://mafft.cbrc.jp/alignment/software/algorithms/algorithms.html')
NEEDLE_DEFAULT_ARGS = 'auto'


def mafft(reads, verbose=False, options=None, threads=None):
    """
    Run a MAFT alignment and return the two sequences.

    @param reads: An iterable of two reads.
    @param verbose: If C{True} print progress info to sys.stderr.
    @param options: A C{str} of options to pass to mafft.
    @return: A C{Reads} instance with the two aligned sequences.
    """
    tempdir = mkdtemp()

    infile = join(tempdir, 'input.fasta')
    out = join(tempdir, 'result.fasta')

    Reads(reads).save(infile)

    if verbose:
        print('Running mafft.', file=sys.stderr)

    Executor().execute("mafft --thread %d %s '%s' > '%s'" % (
        threads, options or '', infile, out))

    # Use 'list' in the following to force reading the FASTA from disk.
    result = Reads(list(FastaReads(out)))
    rmtree(tempdir)

    return result


def needle(reads, verbose=False, options=None):
    """
    Run a Needleman-Wunsch alignment and return the two sequences.

    @param reads: An iterable of two reads.
    @param verbose: If C{True} print progress info to sys.stderr.
    @param options: Additional options to pass to needle.
    @return: A C{Reads} instance with the two aligned sequences.
    """
    tempdir = mkdtemp()

    file1 = join(tempdir, 'file1.fasta')
    with open(file1, 'w') as fp:
        print(reads[0].toString('fasta'), end='', file=fp)

    file2 = join(tempdir, 'file2.fasta')
    with open(file2, 'w') as fp:
        print(reads[1].toString('fasta'), end='', file=fp)

    out = join(tempdir, 'result.fasta')

    def useStderr(e):
        return "Sequences too big. Try 'stretcher'" not in e.stderr

    if verbose:
        print('Running needle.', file=sys.stderr)
    try:
        Executor().execute(
            "needle -asequence '%s' -bsequence '%s' %s "
            "-outfile '%s' -aformat fasta" % (
                file1, file2, options or '', out), useStderr=useStderr)
    except CalledProcessError as e:
        if useStderr(e):
            raise
        else:
            if verbose:
                print('Sequences too long for needle. Falling back to '
                      'stretcher. Be patient!', file=sys.stderr)
            Executor().execute("stretcher -asequence '%s' -bsequence '%s' "
                               "-auto "
                               "-outfile '%s' -aformat fasta" % (
                                   file1, file2, out))

    # Use 'list' in the following to force reading the FASTA from disk.
    result = Reads(list(FastaReads(out)))
    rmtree(tempdir)

    return result


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Compare two sequences.'))

parser.add_argument(
    '--index1', type=int, default=1,
    help='The (1-based) index in the input of the first sequence.')

parser.add_argument(
    '--index2', type=int, default=2,
    help='The (1-based) index in the input of the second sequence.')

parser.add_argument(
    '--align', default=False, action='store_true',
    help=('If given, use mafft (the default) or needle (according to the '
          'algorithm selected by --aligner) to align the two sequences.'))

parser.add_argument(
    '--aligner', default='mafft', choices=('mafft', 'needle'),
    help='The alignment algorithm to use.')

parser.add_argument(
    '--alignerOptions',
    help=('Optional arguments to pass to the alignment algorithm. If the '
          'aligner is mafft, the default options are "%s". If needle, "%s". '
          'Do not try to set the number of threads here - use the --threads '
          'argument instead. If you are using mafft, see %s for some possible '
          'option combinations.' %
          (MAFFT_DEFAULT_ARGS, NEEDLE_DEFAULT_ARGS, MAFFT_ALGORITHMS_URL)))

parser.add_argument(
    '--threads', type=int, default=multiprocessing.cpu_count(),
    help=('The number of threads to use when running the aligner (if --align '
          'is used and the alignment algorithm can make use of multiple '
          'threads (mafft can, needle cannot)).'))

parser.add_argument(
    '--alignmentFile',
    help='The file to save the alignment to (implies --align).')

parser.add_argument(
    '--strict', default=False, action='store_true',
    help='If given, do not allow ambiguous nucleotide symbols to match')

parser.add_argument(
    '--quiet', dest='verbose', default=True, action='store_false',
    help=('Do not print information about aligning, or falling back to '
          'stretcher.'))

parser.add_argument(
    '--noGapLocations', dest='includeGapLocations', action='store_false',
    default=True,
    help='Do not indicate the (1-based) locations of sequence gaps.')

parser.add_argument(
    '--sites',
    help=('Specify (1-based) sequence sites to keep. All other sites will '
          'be ignored. The sites must be given in the form e.g., '
          '24,100-200,260.'))

parser.add_argument(
    '--showDiffs', default=False, action='store_true',
    help='Print (1-based) sites where the sequence nucleotides differ.')

parser.add_argument(
    '--showAmbiguous', default=False, action='store_true',
    help=('Print (1-based) sites where either sequence has an ambiguous '
          'nucleotide code.'))

addFASTACommandLineOptions(parser)
args = parser.parse_args()

keepSequences = set([args.index1 - 1, args.index2 - 1])

reads = list(parseFASTACommandLineOptions(args).filter(
    keepSequences=keepSequences))

if len(reads) == 1:
    if len(keepSequences) == 1:
        # This is ok, they want to compare a sequence with itself.
        reads = Reads([reads[0], reads[0]])
    else:
        print('Could not find both requested sequence indices. Exiting.')
        sys.exit(1)
elif len(reads) != 2:
    print('Could not find both requested sequence indices. Exiting.')
    sys.exit(1)

if args.alignmentFile:
    args.align = True

if args.align:
    len1, len2 = map(len, reads)
    if len1 == len2:
        print('Pre-alignment, sequence lengths were identical: %d' % len1)
    else:
        print('Pre-alignment, sequence lengths: %d, %d (difference %d)' % (
            len1, len2, abs(len1 - len2)))

    if args.aligner == 'mafft':
        # Be careful in examining args.alignerOptions because we want the
        # user to be able to pass an empty string (so check against None
        # before deciding to use the default.)
        options = (MAFFT_DEFAULT_ARGS if args.alignerOptions is None
                   else args.alignerOptions)
        reads = mafft(reads, args.verbose, options=options,
                      threads=args.threads)
    else:
        assert args.aligner == 'needle'
        # Be careful in examining args.alignerOptions because we want the
        # user to be able to pass an empty string (so check against None
        # before deciding to use the default.)
        options = (NEEDLE_DEFAULT_ARGS if args.alignerOptions is None
                   else args.alignerOptions)
        reads = needle(reads, args.verbose, options=options)

    if args.alignmentFile:
        assert reads.save(args.alignmentFile) == 2

offsets = (parseRangeString(args.sites, convertToZeroBased=True)
           if args.sites else None)

read1, read2 = reads
len1, len2 = map(len, reads)
identicalLengths = len1 == len2

# Sanity check.
if args.align:
    assert identicalLengths

match = compareDNAReads(read1, read2, matchAmbiguous=(not args.strict),
                        offsets=offsets)

x = 'Post-alignment, sequence' if args.align else 'Sequence'
if identicalLengths:
    print('%s lengths are identical: %s' % (x, len1))
else:
    print('%s lengths: %d, %d (difference %d)' % (x, len1, len2,
                                                  abs(len1 - len2)))

print(matchToString(match, read1, read2, matchAmbiguous=(not args.strict),
                    offsets=offsets,
                    includeGapLocations=args.includeGapLocations))

if args.showDiffs:
    # Print all sites where the sequences differ.
    width = int(log10(max(len1, len2))) + 1
    headerPrinted = False
    for site, (a, b) in enumerate(zip(read1.sequence, read2.sequence),
                                  start=1):
        if a != b:
            if not headerPrinted:
                print('Differences (site, %s, %s):' % (read1.id, read2.id))
                headerPrinted = True
            print('  %*d %s %s' % (width, site, a, b))

    if not headerPrinted:
        print('No sequence differences found.')

if args.showAmbiguous:
    width = int(log10(max(len1, len2))) + 1
    headerPrinted = False
    for site, (a, b) in enumerate(zip(read1.sequence, read2.sequence),
                                  start=1):
        if len(AMBIGUOUS.get(a, '')) > 1 or len(AMBIGUOUS.get(b, '')) > 1:
            if not headerPrinted:
                print('Ambiguities (site, %s, %s):' % (read1.id, read2.id))
                headerPrinted = True
            print('  %*d %s %s' % (width, site, a, b))

    if not headerPrinted:
        print('No sequence ambiguities found.')



from unittest import TestCase

from dark.codonDistance import codonInformation, findDistance


class TestCodonInformation(TestCase):
    """
    Tests for the codonInformation function.
    """
    def testReturnRightCodon(self):
        """
        Must return the right codon.
        """
        result = codonInformation(['GAT', 'GAC'], ['GCC', 'GCA'])
        self.assertEqual([['GAC', 'GCC']], result[1])
        self.assertEqual([['GAT', 'GCC'], ['GAT', 'GCA'], ['GAC', 'GCA']],
                         result[2])


class TestFindDistance(TestCase):
    """
    Tests for the findDistance function.
    """
    def testFindDistanceThree(self):
        """
        findDistance must return the right distance if codons are different.
        """
        distance = findDistance('ACA', 'TGT')
        self.assertEqual(3, distance)

    def testFindDistanceZero(self):
        """
        findDistance must return the right distance if codons are identical.
        """
        distance = findDistance('ACA', 'ACA')
        self.assertEqual(0, distance)



import sqlite3
from six import string_types
from operator import attrgetter
from cachetools import LRUCache, cachedmethod
from json import dumps
import re
from collections import namedtuple
from os import environ

from dark.database import getDatabaseConnection

TAXONOMY_DATABASE_ENV_VAR = 'DARK_MATTER_TAXONOMY_DATABASE'
TAXONOMY_DATABASE_COMMAND_LINE_OPTION = 'taxonomyDatabase'

LineageElement = namedtuple('LineageElement', ('taxid', 'name', 'rank'))

# These are rank/name tuples that indicate that a lineage is from an RNA
# virus.
RNA_VIRUS_LINEAGE_ELEMENTS = set((
    ('family', 'Retroviridae'),
    ('family', 'Pseudoviridae'),
    ('realm', 'Riboviria'),
))

FUNGUS_ONLY_VIRUS_REGEX = re.compile(
    r'\b(?:mycovirus)\b',
    re.I)

FUNGUS_ONLY_GENERA = {
}

FUNGUS_ONLY_FAMILIES = {
    'Chrysoviridae',  # Infects penicillum. E.g., NC_040738.1
    'Totiviridae',  # Actually infects fungi and protozoa.
}

PLANT_ONLY_VIRUS_REGEX = re.compile(
    r'\b(?:grapevine|mosaic|mottle|blotch virus|viroid|'
    r'tombus|cherry virus|bean endornavirus|coffee ringspot virus|'
    r'wheat stripe virus|Lettuce necrotic yellows virus|citrus yellow|'
    r'Maize fine streak virus|dwarf (fiji)?virus|maize stripe virus|'
    r'chlorotic spot virus)\b',
    re.I)

PLANT_ONLY_GENERA = {
    'Emaravirus',  # In Bunyavirales.
    'Idaeovirus',  # Unassigned.
    'Ourmiavirus',  # In Botourmiaviridae.
    'Polemovirus',  # Unassigned.
    'Sobemovirus',  # Unassigned.
    'Tospovirus',  # (Misclassified?) in Peribunyaviridae (e.g., NC_040742.1)
    'Umbravirus',  # In Tombusviridae.
    'Varicosavirus',  # In Rhabdoviridae.
}

PLANT_ONLY_FAMILIES = {
    'Aspiviridae',  # (Formerly Ophioviridae) in Serpentovirales.
    'Avsunviroidae',  # Viroids.
    'Betaflexiviridae',  # In Tymovirales.
    'Bromoviridae',  # Unassigned.
    'Caulimoviridae',  # In Ortervirales.
    'Closteroviridae',  # Unassigned.
    'Geminiviridae',  # Unassigned.
    'Luteoviridae',  # Unassigned.
    'Nanoviridae',  # Unassigned.
    'Pospiviroidae',  # Unassigned (viroids).
    'Secoviridae',  # In Picornavirales.
    'Tombusviridae',  # Unassigned.
    'Tospoviridae',  # In Bunyavirales.
}


class _UnknownNameError(Exception):
    "An unknown name was used."


class LineageFetcher(object):
    """
    Provide access to the NCBI taxonomy database so we can retrieve the lineage
    of title sequences hit by BLAST.
    """
    def __init__(self, db=None, cursor=None):
        self._db = db or getDatabaseConnection()
        self._cursor = cursor or self._db.cursor()
        self._cache = {}

    def lineage(self, title):
        """
        Get lineage information from the taxonomy database for a given title.

        @param title: A C{str} sequence title (e.g., from a BLAST hit). Of the
            form 'gi|63148399|gb|DQ011818.1| Description...'. It is the gi
            number (63148399 in this example) that is looked up in the taxonomy
            database.
        @return: A C{list} of the taxonomic categories of the title. Each list
            element is an (C{int}, C{str}) 2-tuple, giving a taxonomy id and
            a scientific name. The first element in the list will correspond to
            C{title}, and each successive element is the parent of the
            preceeding one. If no taxonomy is found, the returned list will be
            empty.
        """
        if title in self._cache:
            return self._cache[title]

        lineage = []
        gi = int(title.split('|')[1])
        query = 'SELECT taxID FROM gi_taxid WHERE gi = %d' % gi

        try:
            while True:
                self._cursor.execute(query)
                taxID = self._cursor.fetchone()[0]
                if taxID == 1:
                    break
                query = 'SELECT name FROM names WHERE taxId = %s' % taxID
                self._cursor.execute(query)
                scientificName = self._cursor.fetchone()[0]
                lineage.append((taxID, scientificName))
                # Move up to the parent.
                query = ('SELECT parent_taxID FROM nodes WHERE taxID = %s' %
                         taxID)
        except TypeError:
            lineage = []

        self._cache[title] = lineage
        return lineage

    def close(self):
        """
        Close the database connection and render self invalid. Any subsequent
        re-use of self will raise an error.
        """
        self._cursor.close()
        self._db.close()
        self._cursor = self._db = self._cache = None


class Taxonomy(object):
    """
    Provide access to the NCBI taxonomy database so we can retrieve the
    taxonomy lineage corresponding to an accession number.

    @param dbFilenameOrConnection: Either a C{str} database filename or an
        open sqlite3 database. The database must contain taxonomy information
        with tables and columns named as in the building scripts used in
        https://github.com/acorg/ncbi-taxonomy-database
    """
    CACHE_SIZE = 10E6

    def __init__(self, dbFilenameOrConnection):
        if isinstance(dbFilenameOrConnection, string_types):
            self._db = sqlite3.connect(dbFilenameOrConnection)
            self._closeConnection = True
        else:
            self._db = dbFilenameOrConnection
            self._closeConnection = False
        self._lineageCache = LRUCache(maxsize=self.CACHE_SIZE)
        self._hostsCache = LRUCache(maxsize=self.CACHE_SIZE)
        self._plantVirusCache = LRUCache(maxsize=self.CACHE_SIZE)
        self._fungusVirusCache = LRUCache(maxsize=self.CACHE_SIZE)

    def lineageFromTaxid(self, taxid, skipFunc=None, stopFunc=None):
        """
        Get lineage information from the taxonomy database for a taxonomy id.

        @param taxid: An C{int} taxonomy id.
        @param skipFunc: A function that takes a C{LineageElement} instance
            and returns C{True} if this lineage item should be excluded from
            the returned C{tuple}.
        @param stopFunc: A function that takes a C{LineageElement} instance
            and returns C{True} if the lineage processing should be stopped.
            The final lineage element is still added to the returned value,
            unless a C{skipFunc} is passed that returns C{True} for it.
        @raise ValueError: If the taxonomy id cannot be found in the names
            table or if any taxonomy id cannot be found in the nodes table
            as we move up the hierarchy.
        @raise AttributeError: If C{self.close} has been called.
        @return: A C{tuple} of the taxonomic categories of the title. Each
            tuple element is a C{LineageElement}, giving a taxonomy id a
            (scientific) name, and the rank (species, genus, etc). The first
            element in the list corresponds to the passed C{taxid} and each
            successive element is the parent of the preceeding one. The top
            level of the taxonomy hierarchy (with taxid = 1) is not included
            in the returned list. Note that the returned C{tuple} may be empty,
            depending on C{skipFunc} and C{stopFunc}.
        """
        cursor = self._db.cursor()
        execute = cursor.execute
        fetchone = cursor.fetchone

        lineage = []

        while taxid != 1:
            execute('SELECT name FROM names WHERE taxid = ?', (taxid,))
            result = fetchone()
            if result is None:
                raise ValueError(
                    'Could not find taxonomy id %r in names table' %
                    (taxid,))
            name = result[0]
            execute('SELECT parent_taxid, rank FROM nodes WHERE taxid = ?',
                    (taxid,))
            result = fetchone()
            if result is None:
                raise ValueError(
                    'Could not find taxonomy id %r in nodes table' %
                    (taxid,))

            parentTaxid, rank = result
            element = LineageElement(taxid, name, rank)

            if not (skipFunc and skipFunc(element)):
                lineage.append(element)

            if stopFunc and stopFunc(element):
                break

            taxid = parentTaxid

        return tuple(lineage)

    @cachedmethod(attrgetter('_lineageCache'))
    def lineage(self, id_, skipFunc=None, stopFunc=None):
        """
        Get lineage information from the taxonomy database for an
        accession number, name, or taxonomy id.

        @param id_: A C{str} accession number, C{str} name (e.g.,
            'Hymenoptera'), or an C{int} taxonomy id. If a C{str} accession
            number, C{id_} must include the version (e.g., the ".1" part of
            "DQ011818.1") if the taxonomy database in use uses it (as is the
            case with the database built by
            https://github.com/acorg/ncbi-taxonomy-database)
        @param skipFunc: A function that takes a C{LineageElement} instance
            and returns C{True} if this lineage item should be excluded from
            the returned C{tuple}.
        @param stopFunc: A function that takes a C{LineageElement} instance
            and returns C{True} if the lineage processing should be stopped.
            The final lineage element is still added to the returned value,
            unless a C{skipFunc} is passed that returns C{True} for it.
        @raise ValueError: If a taxonomy id (or one of its higher-level
            taxonomy ids) cannot be found.
        @raise AttributeError: If C{self.close} has been called.
        @return: A C{tuple} of the taxonomic categories of the title. Each
            tuple element is a C{LineageElement}, giving a taxonomy id a
            (scientific) name, and the rank (species, genus, etc). The first
            element in the tuple corresponds to the passed C{id_} and each
            successive element is the parent of the preceeding one. The top
            level of the taxonomy hierarchy (with taxid = 1) is not included
            in the returned tuple.
        """
        if isinstance(id_, int):
            return self.lineageFromTaxid(
                id_, skipFunc=skipFunc, stopFunc=stopFunc)

        cursor = self._db.cursor()

        cursor.execute(
            'SELECT taxid FROM accession_taxid WHERE accession = ?', (id_,))
        row = cursor.fetchone()
        if row:
            return self.lineageFromTaxid(
                int(row[0]), skipFunc=skipFunc, stopFunc=stopFunc)

        try:
            taxid = self.taxIdFromName(id_)
        except _UnknownNameError:
            raise ValueError(
                'Could not find %r in accession_taxid or names tables' % id_)
        else:
            return self.lineageFromTaxid(
                taxid, skipFunc=skipFunc, stopFunc=stopFunc)

    @cachedmethod(attrgetter('_hostsCache'))
    def hostsFromTaxid(self, taxid):
        """
        Get host information from the taxonomy database for a taxonomy id.

        @param taxid: An C{int} taxonomy id.
        @raise AttributeError: If C{self.close} has been called.
        @return: A C{set} of C{str} host names. If no host information is found
            for C{taxid}, return C{None}.
        """
        cursor = self._db.cursor()
        cursor.execute('SELECT hosts FROM hosts WHERE taxid = ?', (taxid,))
        row = cursor.fetchone()
        if row:
            return set(row[0].split(','))

    def taxIdFromName(self, name):
        """
        Get a taxonomy id from a name.

        @param name: A C{str} taxonomy name. This must be a unique name, as
            found in the names.dmp file of the NCBI taxonomy names.dmp file.
        @raise _UnknownNameError: If a taxonomy id cannot be found for C{name}.
        @return: An C{int} taxonomy id.
        """
        cursor = self._db.cursor()
        cursor.execute('SELECT taxid FROM names WHERE name = ?', (name,))
        row = cursor.fetchone()
        if row:
            return int(row[0])

        raise _UnknownNameError('Name %r not found in names tables' % name)

    def hosts(self, id_):
        """
        Get host information from the taxonomy database for an accession number
        or taxonomy id.

        @param id_: A C{str} accession number or C{int} taxonomy id.
            If a C{str}, C{id_} must include the version (e.g., the
            ".1" part of "DQ011818.1") if the taxonomy database in use uses
            it (as is the case with the database built by
            https://github.com/acorg/ncbi-taxonomy-database
        @raise AttributeError: If C{self.close} has been called.
        @return: A C{set} of C{str} host names, or C{None} if no host
            information.
        """
        if isinstance(id_, int):
            return self.hostsFromTaxid(id_)

        cursor = self._db.cursor()

        cursor.execute(
            'SELECT taxid FROM accession_taxid WHERE accession = ?',
            (id_,))
        row = cursor.fetchone()
        if row:
            return self.hostsFromTaxid(int(row[0]))

        cursor.execute('SELECT taxid FROM names WHERE name = ?', (id_,))
        row = cursor.fetchone()
        if row:
            return self.hostsFromTaxid(int(row[0]))

    @cachedmethod(attrgetter('_fungusVirusCache'))
    def isFungusOnlyVirus(self, lineage, title=None):
        """
        Determine whether a lineage corresponds to a fungus-only virus (i.e.,
        a virus that only infects fungi hosts).

        @param lineage: An iterable of C{tuple}s of taxonomy id, scientific
            name, and rank as returned by C{Taxonomy.lineage}.
        @param title: If not C{None}, a C{str} title of the virus.
        @return: C{True} if the lineage corresponds to a fungus-only virus,
            C{False} otherwise.
        """
        for taxid, name, rank in lineage:
            if (rank == 'family' and name in FUNGUS_ONLY_FAMILIES or
                    rank == 'genus' and name in FUNGUS_ONLY_GENERA):
                return True

        if title and FUNGUS_ONLY_VIRUS_REGEX.search(title):
            return True

        # Do host taxonomy database lookups. Try to look up host
        # information at all taxonomy levels. This is (of course) slower,
        # but it (likely, as with plant viruses) results in more
        # fungus-only viruses being identified.
        for taxid, _, _ in lineage:
            if self.hosts(taxid) == {'fungi'}:
                return True

        return False

    @cachedmethod(attrgetter('_plantVirusCache'))
    def isPlantOnlyVirus(self, lineage, title=None):
        """
        Determine whether a lineage corresponds to a plant-only virus (i.e.,
        a virus that only infects plant hosts).

        @param lineage: An iterable of C{tuple}s of taxonomy id, scientific
            name, and rank as returned by C{Taxonomy.lineage}.
        @param title: If not C{None}, a C{str} title of the virus.
        @return: C{True} if the lineage corresponds to a plant-only virus,
            C{False} otherwise.
        """
        for taxid, name, rank in lineage:
            if (rank == 'family' and name in PLANT_ONLY_FAMILIES or
                    rank == 'genus' and name in PLANT_ONLY_GENERA):
                return True

        if title and PLANT_ONLY_VIRUS_REGEX.search(title):
            return True

        # Do host taxonomy database lookups. Try to look up host information at
        # all taxonomy levels. This is (of course) slower, but it does result
        # in more plant-only viruses being identified, e.g., with rank
        # 'unclassified' and name 'Partitiviridae'.
        for taxid, _, _ in lineage:
            if self.hosts(taxid) == {'plants'}:
                return True

        return False

    @staticmethod
    def subsetLineageByRanks(lineage, func):
        """
        Extract certain ranks from a lineage.

        @param lineage: An iterable of C{tuple}s of taxonomy id, scientific
            name, and rank as returned by C{Taxonomy.lineage}.
        @param func: A function that accepts a C{str} rank (which may be '-')
            and returns C{True} or C{False} according to whether the lineage
            element should be retained.
        @return: A filter object that yields C{tuple}s of taxonomy id,
            scientific name, and rank, where the elements are those that
            C{func} returns C{True}. The elements in the returned value will
            be in the order in which they are present in C{lineage}.
        """
        return filter(lambda l: func(l.rank), lineage)

    def close(self):
        """
        Close the database connection (if we opened it).
        """
        if self._closeConnection:
            self._db.close()
        # Set self._db to None so self.lineage will raise an AttributeError
        # exception if it is called again.
        self._db = None

    def __enter__(self):
        return self

    def __exit__(self, excType, excValue, traceback):
        self.close()


def isRetrovirus(lineage):
    """
    Determine whether a lineage corresponds to a retrovirus.

    @param lineage: An iterable of C{LineageElement} instances.
    @return: C{True} if the lineage corresponds to a retrovirus, C{False}
        otherwise.
    """
    for l in lineage:
        if l.rank == 'family' and l.name == 'Retroviridae':
            return True

    return False


def isRNAVirus(lineage):
    """
    Determine whether a lineage corresponds to an RNA virus.

    @param lineage: An iterable of C{LineageElement} instances.
    @return: C{True} if the lineage corresponds to an RNA virus, C{False}
        otherwise.
    """
    for l in lineage:
        if (l.rank, l.name) in RNA_VIRUS_LINEAGE_ELEMENTS:
            return True

    return False


def isDNAVirus(lineage):
    """
    Determine whether a lineage corresponds to an DNA virus.

    @param lineage: An iterable of C{LineageElement} instances.
    @return: C{True} if the lineage corresponds to a DNA virus, C{False}
        otherwise.
    """
    return not isRNAVirus(lineage)


def formatLineage(lineage, namesOnly=False, separator=None, prefix=''):
    """
    Format a lineage for printing.

    @param lineage: An iterable of C{LineageElement} instances.
    @param namesOnly: If C{True} only print taxonomic names.
    @param separator: A C{str} separator to put between fields. If C{None},
        return a space-padded aligned columns.
    @param prefix: A C{str} to put at the start of each line.
    @return: A formatted C{str} for printing.
    """
    if namesOnly:
        # The separator is guaranteed to be set by our caller.
        return prefix + separator.join(l.name for l in lineage)

    if separator is not None:
        return '\n'.join(
            '%s%s%s%s%s%s' % (prefix, rank, separator, name, separator, taxid)
            for (taxid, name, rank) in lineage)

    # This is a bit slow, walking through the lineage list 3 times.
    taxidWidth = max(len(str(l.taxid)) for l in lineage)
    nameWidth = max(len(l.name) for l in lineage)
    rankWidth = max(len(l.rank) for l in lineage)

    return '\n'.join(
        '%s%-*s %-*s %*d' % (
            prefix, rankWidth, rank, nameWidth, name, taxidWidth, taxid)
        for (taxid, name, rank) in lineage)


def lineageTaxonomyLinks(lineage):
    """
    Get HTML links for a lineage.

    @param lineage: An iterable of C{LineageElement} instances.
    @return: A C{list} of HTML C{str} links.
    """
    names = [l.name for l in lineage]
    assert names[-1] == 'Viruses'
    names[0] = 'taxon'

    taxids = [l.taxid for l in lineage]

    return [
        '<a href="%s%s">%s</a>' % (
            'https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=',
            taxid, name) for taxid, name in list(zip(taxids, names))[:-1]
    ]


class _HierarchyNode(object):
    def __init__(self, name):
        self.name = name
        self.count = 0
        self.nodes = {}
        self.tips = []

    def addChild(self, name):
        self.count += 1
        try:
            node = self.nodes[name]
        except KeyError:
            node = self.nodes[name] = _HierarchyNode(name)

        return node

    def addTip(self, name, genomeAccession):
        self.count += 1
        self.tips.append((name, genomeAccession))

    def toDict(self):
        nodes = [node.toDict() for node in self.nodes.values()]

        if self.tips:
            nodes.extend([{
                'text': name,
                'href': '#pathogen-' + accession,
            } for (name, accession) in sorted(self.tips)])

        return {
            # 'count': self.count,
            'nodes': nodes,
            # 'name': self.name,
            # 'tips': self.tips,
            'text': '%s (%d)' % (self.name, self.count),
        }


class Hierarchy(object):
    """
    Collect information about a lineages in a taxonomy hierarchy.
    """
    def __init__(self, rootName='Viruses'):
        self._root = _HierarchyNode(rootName)

    def add(self, lineage, genomeAccession):
        """
        Add a new lineage.

        @param lineage: An iterable of C{LineageElement} instances.
        @param genomeAccession: A C{str} pathogen accession number.
        """
        names = [l.name for l in lineage]

        assert names[-1] == 'Viruses'

        n = len(names) - 2
        node = self._root

        while n:
            node = node.addChild(names[n])
            n -= 1

        node.addTip(names[0], genomeAccession)

    def toJSON(self):
        return dumps([self._root.toDict()])


def addTaxonomyDatabaseCommandLineOptions(parser):
    """
    Add standard taxonomy database command-line options to an argparse parser.

    @param parser: An C{argparse.ArgumentParser} instance.
    """
    parser.add_argument(
        '--' + TAXONOMY_DATABASE_COMMAND_LINE_OPTION, metavar='DATABASE-FILE',
        help=('The file holding an sqlite3 taxonomy database. See '
              'https://github.com/acorg/ncbi-taxonomy-database for how to '
              'build one. If not specified, the value in the %r environment '
              'variable (if any) will be used.' % TAXONOMY_DATABASE_ENV_VAR), )


def parseTaxonomyDatabaseCommandLineOptions(args, parser):
    """
    Examine parsed command-line options and return a Taxonomy instance. Exits
    if no taxonomy database is given or named in the TAXONOMY_DATABASE_ENV_VAR
    environment variable.

    @param args: An argparse namespace, as returned by the argparse
        C{parse_args} function.
    @param parser: An C{argparse.ArgumentParser} instance.
    @return: A C{Taxonomy} instance, or C{None} if no database filename is
        given or set via the environment variable named in
        TAXONOMY_DATABASE_ENV_VAR.
    """
    filename = args.taxonomyDatabase or environ.get(
        TAXONOMY_DATABASE_ENV_VAR)

    if filename:
        return Taxonomy(filename)
    else:
        parser.error(
            'A taxonomy database file must be given, either via --%s on '
            'the command line or via the %r environment variable.' %
            (TAXONOMY_DATABASE_COMMAND_LINE_OPTION, TAXONOMY_DATABASE_ENV_VAR))



from unittest import TestCase

from dark.hsp import HSP, LSP


class TestHSP(TestCase):
    """
    Tests of the L{dark.hsp.HSP} class.
    """

    def testExpectedAttributes(self):
        """
        An HSP must have the expected attributes.
        """
        hsp = HSP(7, readStart=1, readEnd=2,
                  readStartInSubject=3, readEndInSubject=4,
                  subjectStart=5, subjectEnd=6,
                  readMatchedSequence='aaa', subjectMatchedSequence='ccc',
                  readFrame=8, subjectFrame=9)
        self.assertEqual(1, hsp.readStart)
        self.assertEqual(2, hsp.readEnd)
        self.assertEqual(3, hsp.readStartInSubject)
        self.assertEqual(4, hsp.readEndInSubject)
        self.assertEqual(5, hsp.subjectStart)
        self.assertEqual(6, hsp.subjectEnd)
        self.assertEqual('aaa', hsp.readMatchedSequence)
        self.assertEqual('ccc', hsp.subjectMatchedSequence)
        self.assertEqual(7, hsp.score.score)
        self.assertEqual(8, hsp.readFrame)
        self.assertEqual(9, hsp.subjectFrame)

    def testEqual(self):
        """
        Two HSPs must compare properly with ==
        """
        self.assertEqual(HSP(7), HSP(7))

    def testLt(self):
        """
        Two HSPs must compare properly with <
        """
        self.assertTrue(HSP(7) < HSP(8))

    def testBetterThanTrue(self):
        """
        Two HSP scores must compare properly with betterThan if the passed
        score is worse than the score of the HSP.
        """
        self.assertTrue(HSP(7).betterThan(5))

    def testBetterThanFalse(self):
        """
        Two HSP scores must compare properly with betterThan if the passed
        score is better than the score of the HSP.
        """
        self.assertFalse(HSP(5).betterThan(7))

    def testToDict(self):
        """
        The toDict method must return the expected dictionary.
        """
        hsp = HSP(0, readStart=1, readEnd=2,
                  readStartInSubject=3, readEndInSubject=4,
                  subjectStart=5, subjectEnd=6,
                  readMatchedSequence='aaa', subjectMatchedSequence='ccc',
                  readFrame=7, subjectFrame=8, identicalCount=9,
                  percentIdentical=99.3, positiveCount=10, percentPositive=3.0)

        self.assertEqual(
            {
                'score': 0,
                'readStart': 1,
                'readEnd': 2,
                'readStartInSubject': 3,
                'readEndInSubject': 4,
                'subjectStart': 5,
                'subjectEnd': 6,
                'readFrame': 7,
                'subjectFrame': 8,
                'identicalCount': 9,
                'percentIdentical': 99.3,
                'positiveCount': 10,
                'percentPositive': 3.0,
                'readMatchedSequence': 'aaa',
                'subjectMatchedSequence': 'ccc',
            },
            hsp.toDict())


class TestLSP(TestCase):
    """
    Tests of the L{dark.hsp.blast.LSP} class.
    """

    def testExpectedAttributes(self):
        """
        An LSP must have the expected attributes.
        """
        lsp = LSP(7, readStart=1, readEnd=2,
                  readStartInSubject=3, readEndInSubject=4,
                  subjectStart=5, subjectEnd=6,
                  readMatchedSequence='aaa', subjectMatchedSequence='ccc',
                  readFrame=8, subjectFrame=9)
        self.assertEqual(1, lsp.readStart)
        self.assertEqual(2, lsp.readEnd)
        self.assertEqual(3, lsp.readStartInSubject)
        self.assertEqual(4, lsp.readEndInSubject)
        self.assertEqual(5, lsp.subjectStart)
        self.assertEqual(6, lsp.subjectEnd)
        self.assertEqual('aaa', lsp.readMatchedSequence)
        self.assertEqual('ccc', lsp.subjectMatchedSequence)
        self.assertEqual(7, lsp.score.score)
        self.assertEqual(8, lsp.readFrame)
        self.assertEqual(9, lsp.subjectFrame)

    def testEqual(self):
        """
        Two LSPs must compare properly with ==
        """
        self.assertEqual(LSP(7), LSP(7))

    def testLt(self):
        """
        Two LSPs must compare properly with <
        """
        self.assertTrue(LSP(8) < LSP(7))

    def testBetterThanTrue(self):
        """
        Two LSP scores must compare properly with betterThan if the passed
        score is worse than the score of the LSP.
        """
        self.assertTrue(LSP(5).betterThan(7))

    def testBetterThanFalse(self):
        """
        Two LSP scores must compare properly with betterThan if the passed
        score is better than the score of the LSP.
        """
        self.assertFalse(LSP(7).betterThan(5))

    def testToDict(self):
        """
        The toDict method must return the expected dictionary.
        """
        lsp = LSP(0, readStart=1, readEnd=2,
                  readStartInSubject=3, readEndInSubject=4,
                  subjectStart=5, subjectEnd=6,
                  readMatchedSequence='aaa', subjectMatchedSequence='ccc',
                  readFrame=7, subjectFrame=8, identicalCount=9,
                  percentIdentical=99.3, positiveCount=10,
                  percentPositive=9.9)

        self.assertEqual(
            {
                'score': 0,
                'readStart': 1,
                'readEnd': 2,
                'readStartInSubject': 3,
                'readEndInSubject': 4,
                'subjectStart': 5,
                'subjectEnd': 6,
                'readFrame': 7,
                'subjectFrame': 8,
                'identicalCount': 9,
                'percentIdentical': 99.3,
                'positiveCount': 10,
                'percentPositive': 9.9,
                'readMatchedSequence': 'aaa',
                'subjectMatchedSequence': 'ccc',
            },
            lsp.toDict())



#!/usr/bin/env python

from __future__ import print_function

import sys
import argparse
from itertools import chain

from dark.civ.proteins import SqliteIndex
from dark.errors import NoSuchGenomeError
from dark.filter import (
    addFASTAFilteringCommandLineOptions, parseFASTAFilteringCommandLineOptions)
from dark.genbank import GenomeRanges
from dark.genomes import GenomeProteinInfo
from dark.reads import Reads
from dark.sam import SAMFilter, samReferences
from dark.utils import pct


def summarize(gpi, sortOn, minReadOffsetCount):
    """
    Print a summary of the genome proteins.

    @param gpi: A C{GenomeProteinInfo} instance.
    @param sortOn: How to sort proteins for output. One of 'coverage',
        'depth', 'name', 'offset', or 'readCount'.
    @param minReadOffsetCount: The minimum number of reads offsets that must
        overlap a protein for the read to be considered as sufficiently
        intersecting the protein.
    """
    genome = gpi.genome

    print('Summary of %s (%s):' % (genome['name'], genome['accession']))
    print('  Length: %d' % genome['length'])
    print('  Protein count: %d' % genome['proteinCount'])
    print('  Total protein offsets: %s' % (
        pct(len(gpi.offsets), genome['length'])))

    if gpi.samFiles:
        print('  SAM files analyzed: %d' % len(gpi.samFiles))
        for i, filename in enumerate(gpi.samFiles, start=1):
            print('    %d: %s' % (i, filename))
    else:
        return

    print('  Whole genome coverage (not just proteins):')
    print('    Reads matching genome: %d' % len(gpi.readIdsMatchingGenome))
    print('    Covered genome offsets: %s' % (
        pct(len(gpi.coveredOffsetCount), genome['length'])))
    print('    Average depth across genome: %.3f' % (
        sum(gpi.coveredOffsetCount.values()) / genome['length']))

    coveredProteinOffsetCount = coveredProteinBasesCount = 0
    for offset in gpi.offsets:
        coveredProteinOffsetCount += bool(gpi.coveredOffsetCount[offset])
        coveredProteinBasesCount += gpi.coveredOffsetCount[offset]

    print('  Total protein coverage (irrespective of minReadOffsetCount):')
    print('    Reads matching proteins: %d' % len(gpi.readIdsForAllProteins()))
    print('    Proteins with any coverage: %s' %
          pct(len(gpi.coveredProteins), genome['proteinCount']))
    print('    Covered protein offsets: %s' % (
        pct(coveredProteinOffsetCount, len(gpi.offsets))))
    print('    Average depth across proteins: %.3f' % (
        coveredProteinBasesCount / len(gpi.offsets)))

    if sortOn == 'name':
        def key(proteinAccession):
            return gpi.proteins[proteinAccession]['name']
        reverse = False
    elif sortOn == 'offset':
        def key(proteinAccession):
            return GenomeRanges(
                gpi.proteins[proteinAccession]['offsets']).ranges[0][0]
        reverse = False
    elif sortOn == 'readCount':
        def key(proteinAccession):
            coverage = gpi.proteinCoverageInfo(proteinAccession)
            return len(coverage['readIds'])
        reverse = True
    elif sortOn == 'coverage':
        def key(proteinAccession):
            coverage = gpi.proteinCoverageInfo(proteinAccession)
            return coverage['coveredOffsets'] / coverage['ntLength']
        reverse = True
    elif sortOn == 'depth':
        def key(proteinAccession):
            coverage = gpi.proteinCoverageInfo(proteinAccession)
            return coverage['totalBases'] / coverage['ntLength']
        reverse = True

    if minReadOffsetCount is None:
        print('  Proteins covered (no minReadOffsetCount):')
    else:
        print('  Proteins covered (minReadOffsetCount=%d):' %
              minReadOffsetCount)

    proteinCount = 0
    for proteinAccession in sorted(gpi.coveredProteins, key=key,
                                   reverse=reverse):
        protein = gpi.proteins[proteinAccession]

        coverage = gpi.proteinCoverageInfo(proteinAccession,
                                           minReadOffsetCount)

        readCount = len(coverage['readIds'])

        if readCount:
            proteinCount += 1

            print('    %d: %s (%d AA, %d nt with stop codon, %s)' %
                  (proteinCount, protein['product'], protein['length'],
                   protein['length'] * 3 + 3, protein['accession']))

            print('      Read count: %d' % readCount)

            print('      Covered offsets: %s' % (
                pct(coverage['coveredOffsets'], coverage['ntLength'])))

            print('      Average depth: %.3f' % (
                coverage['totalBases'] / coverage['ntLength']))

            print('      Offsets: %s' % protein['offsets'])

    print('  Proteins matched: %s (sorted by %s):' % (
        pct(proteinCount, genome['proteinCount']), sortOn))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Print SAM/BAM file protein match statistics.')

    parser.add_argument(
        '--proteinGenomeDatabase', required=True,
        help=('The filename of an Sqlite3 database holding protein and '
              'genome information, as built by make-protein-database.py'))

    parser.add_argument(
        '--progress', default=False, action='store_true',
        help='Print progress info to standard error.')

    parser.add_argument(
        '--sortOn', default='readCount',
        choices=('coverage', 'depth', 'name', 'offset', 'readCount'),
        help='How to sort proteins for output.')

    parser.add_argument(
        '--minReadOffsetCount', type=int,
        help=('The minimum number of reads offsets that must overlap a '
              'protein for the read to be considered as sufficiently '
              'intersecting the protein. Use this to prevent reads that '
              'just overlap the protein in a very small number offsets '
              'from being counted.'))

    parser.add_argument(
        '--skipTranslationChecks', dest='checkTranslations',
        action='store_false', default=True,
        help=('Skip the sanity check that database protein sequences can all '
              'be translated from the database genome sequence.'))

    addFASTAFilteringCommandLineOptions(parser)
    SAMFilter.addFilteringOptions(
        parser, samfileIsPositional=False, samfileAction='append',
        samfileNargs='*', samfileRequired=False)

    args = parser.parse_args()

    samfiles = list(chain.from_iterable(args.samfile)) if args.samfile else []

    if samfiles:
        if args.referenceId:
            referenceIds = args.referenceId
        else:
            # If all SAM files have just one reference and they're all the
            # same, use that. Else complain.
            referenceIds = set()
            for filename in samfiles:
                referenceIds.update(samReferences(filename))

            if len(referenceIds) != 1:
                print('No reference id(s) specified with --referenceId, and '
                      'the given SAM/BAM files do not contain exactly one '
                      '(identical) reference. Please use --referenceId')
                sys.exit(1)
    else:
        if args.referenceId:
            referenceIds = args.referenceId
        else:
            print('No reference id(s) specified with --referenceId.')
            sys.exit(1)

    # We don't have a file of reads, we just want a read filter that we
    # can use to filter the SAM file query sequences.
    reads = parseFASTAFilteringCommandLineOptions(args, Reads())
    samFilter = SAMFilter.parseFilteringOptions(args, reads.filterRead)
    filterAlignment = samFilter.filterAlignment

    proteinGenomeDB = SqliteIndex(args.proteinGenomeDatabase)

    for referenceId in referenceIds:
        try:
            gpInfo = GenomeProteinInfo(
                referenceId, proteinGenomeDB,
                checkTranslations=args.checkTranslations)
        except NoSuchGenomeError:
            print('Reference %r not found in genome database. Ignoring.' %
                  referenceId, file=sys.stderr)
        else:
            if samfiles:
                if args.progress:
                    print('Processing %d SAM file%s for matches with %r:' %
                          (len(samfiles), '' if len(samfiles) == 1 else 's',
                           referenceId), file=sys.stderr)
                for i, filename in enumerate(samfiles, start=1):
                    if args.progress:
                        print('  %d: %s' % (i, filename), file=sys.stderr)
                    gpInfo.addSAM(filename, filterAlignment)

            summarize(gpInfo, args.sortOn, args.minReadOffsetCount)



from __future__ import division

import os
import string
import six
import bz2
import gzip
from os.path import basename
from contextlib import contextmanager
from re import compile
import numpy as np


def numericallySortFilenames(names):
    """
    Sort (ascending) a list of file names by their numerical prefixes.
    The number sorted on is the numeric prefix of the basename of
    the given filename. E.g., '../output/1.json.bz2' will sort before
    '../output/10.json.bz2'.

    @param: A C{list} of file names, each of whose basename starts with a
        string of digits.
    @return: The sorted C{list} of full file names.
    """

    def numericPrefix(name):
        """
        Find any numeric prefix of C{name} and return it as an C{int}.

        @param: A C{str} file name, whose name possibly starts with digits.
        @return: The C{int} number at the start of the name, else 0 if
            there are no leading digits in the name.
        """
        count = 0
        for ch in name:
            if ch in string.digits:
                count += 1
            else:
                break
        return 0 if count == 0 else int(name[0:count])

    return sorted(names, key=lambda name: numericPrefix(basename(name)))


# Set up a median function that works on different Python versions and on pypy
# (even if numpypy is still broken).
#
# First try numpy, and check that an ndarray has a partition method (to avoid a
# shortcoming in pypy numpy, see
# https://bitbucket.org/pypy/numpy/issues/12/median-calling-ndarraypartition
# If numpy seems ok, use it. Else try for the Python 3.4 median function, else
# write our own.

_a = np.array([1])

try:
    _a.partition
except AttributeError:
    # Cannot use numpy. Try for the new (Python 3.4) built-in function.
    try:
        from statistics import median as _median
    except ImportError:
        def _median(numbers):
            """
            Find the median of a set of numbers.

            @param numbers: A C{list} of numeric values.
            @return: The median value in C{numbers}.
            """
            numbers = sorted(numbers)
            n = len(numbers)
            if n % 2:
                return numbers[n // 2]
            else:
                n //= 2
                return (numbers[n - 1] + numbers[n]) / 2
else:
    # We have a working numpy. Just make sure we raise on an empty list as
    # numpy returns numpy.nan
    _median = np.median

del _a


def median(l):
    """
    Find the median of a set of numbers.

    @param l: A C{list} of numeric values.
    @raise ValueError: If C{l} is empty.
    @return: The median value in C{l}.
    """

    # Handle the zero length case ahead of time because various median
    # implementations do different things. numpy returns nan, whereas Python
    # 3.4 and later raise StatisticsError

    if len(l) == 0:
        # Match the empty-argument exception message of Python's built-in max
        # and min functions.
        raise ValueError('arg is an empty sequence')
    else:
        return _median(l)


@contextmanager
def asHandle(fileNameOrHandle, mode='rt', encoding='UTF-8'):
    """
    Decorator for file opening that makes it easy to open compressed files
    and which can be passed an already-open file handle or a file name.
    Based on L{Bio.File.as_handle}.

    @param fileNameOrHandle: Either a C{str} or a file handle.
    @param mode: The C{str} mode to use for opening the file.
    @param encoding: The C{str} encoding to use when opening the file.
    @return: A generator that can be turned into a context manager via
        L{contextlib.contextmanager}.
    """
    if isinstance(fileNameOrHandle, six.string_types):
        if fileNameOrHandle.endswith('.gz'):
            if six.PY3:
                yield gzip.open(fileNameOrHandle, mode=mode, encoding=encoding)
            else:
                yield gzip.GzipFile(fileNameOrHandle, mode=mode)
        elif fileNameOrHandle.endswith('.bz2'):
            if six.PY3:
                yield bz2.open(fileNameOrHandle, mode=mode, encoding=encoding)
            else:
                yield bz2.BZ2File(fileNameOrHandle, mode=mode)
        else:
            # Putting mode=mode, encoding=encoding into the following
            # causes a hard-to-understand error from the mocking library.
            with open(fileNameOrHandle) as fp:
                yield fp
    else:
        yield fileNameOrHandle


_rangeRegex = compile(r'^\s*(\d+)(?:\s*-\s*(\d+))?\s*$')


def parseRangeString(s, convertToZeroBased=False):
    """
    Parse a range string of the form 1-5,12,100-200.

    @param s: A C{str} specifiying a set of numbers, given in the form of
        comma separated numeric ranges or individual indices.
    @param convertToZeroBased: If C{True} all indices will have one
        subtracted from them.
    @return: A C{set} of all C{int}s in the specified set.
    """

    result = set()
    for _range in s.split(','):
        match = _rangeRegex.match(_range)
        if match:
            start, end = match.groups()
            start = int(start)
            if end is None:
                end = start
            else:
                end = int(end)
            if start > end:
                start, end = end, start
            if convertToZeroBased:
                result.update(range(start - 1, end))
            else:
                result.update(range(start, end + 1))
        else:
            raise ValueError(
                'Illegal range %r. Ranges must single numbers or '
                'number-number.' % _range)

    return result


if six.PY3:
    from six import StringIO
else:
    from six import StringIO as sixStringIO

    class StringIO(sixStringIO):
        """
        A StringIO class that can be used as a context manager, seeing as the
        six.StringIO class does not provide __enter__ and __exit__ methods
        under Python 2.
        """
        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.close()
            return False


def baseCountsToStr(counts):
    """
    Convert base counts to a string.

    @param counts: A C{Counter} instance.
    @return: A C{str} representation of nucleotide counts at an offset.
    """
    return ' '.join([
        ('%s:%d' % (base, counts[base])) for base in sorted(counts)])


def nucleotidesToStr(nucleotides, prefix=''):
    """
    Convert offsets and base counts to a string.

    @param nucleotides: A C{defaultdict(Counter)} instance, keyed
        by C{int} offset, with nucleotides keying the Counters.
    @param prefix: A C{str} to put at the start of each line.
    @return: A C{str} representation of the offsets and nucleotide
        counts for each.
    """
    result = []
    for offset in sorted(nucleotides):
        result.append(
            '%s%d: %s' % (prefix, offset,
                          baseCountsToStr(nucleotides[offset])))
    return '\n'.join(result)


def countPrint(mesg, count, len1, len2=None):
    """
    Format a message followed by an integer count and a percentage (or
    two, if the sequence lengths are unequal).

    @param mesg: a C{str} message.
    @param count: a numeric value.
    @param len1: the C{int} length of sequence 1.
    @param len2: the C{int} length of sequence 2. If C{None}, will
        default to C{len1}.
    @return: A C{str} for printing.
    """
    def percentage(a, b):
        """
        What percent of a is b?

        @param a: a numeric value.
        @param b: a numeric value.
        @return: the C{float} percentage.
        """
        return 100.0 * a / b if b else 0.0

    if count == 0:
        return '%s: %d' % (mesg, count)
    else:
        len2 = len2 or len1
        if len1 == len2:
            return '%s: %d/%d (%.2f%%)' % (
                mesg, count, len1, percentage(count, len1))
        else:
            return ('%s: %d/%d (%.2f%%) of sequence 1, '
                    '%d/%d (%.2f%%) of sequence 2' % (
                        mesg,
                        count, len1, percentage(count, len1),
                        count, len2, percentage(count, len2)))


def pct(a, b):
    """
    Format a string showing two integers and what percentage the first
    is of the second.

    @param a: An C{int}, the numerator.
    @param b: An C{int}, the denominator.
    """
    assert 0 <= a <= b
    if b:
        return ('%d/%d (%.3f%%)' %
                (a, b, (a / b if b else 0.0) * 100.0))
    else:
        return '0/0 (0.000%)'


@contextmanager
def cd(newdir):
    """
    Trivial context manager for temporarily switching directory.

    @param dir: A C{str} directory to cd to.
    """
    olddir = os.getcwd()
    try:
        os.chdir(newdir)
        yield newdir
    finally:
        os.chdir(olddir)



#!/usr/bin/env python

from __future__ import print_function

from random import choice
from operator import itemgetter

from dark.aa import CODONS
from dark.reads import (
    Read, addFASTACommandLineOptions, parseFASTACommandLineOptions)


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description=('Given amino acid FASTA on stdin, write FASTA with DNA '
                     'sequences to stdout, where the DNA sequences would '
                     'translate to the given amino acid seqeunces. The DNA '
                     'bases used for each amino acid is either fixed or is '
                     'chosen at random from the corresponding codons.'))

    parser.add_argument(
        '--random', action='store_true',
        help=('Use codons chosen at random for each amino acid. Otherwise a '
              'fixed codon will be used for each amino acid so that the '
              'output for a given amino acid sequence is always identical.'))

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)

    getCodon = choice if args.random else itemgetter(0)

    for aaRead in reads:
        dnaRead = Read(aaRead.id,
                       ''.join(getCodon(CODONS[aa]) for aa in aaRead.sequence))
        print(dnaRead.toString('fasta'), end='')



from __future__ import division

from dark.utils import countPrint
from dark.reads import DNAKozakRead
try:
    from itertools import zip_longest
except ImportError:
    # zip_longest does not exist in Python 2.7 itertools. We should be able
    # to get it via from six.moves import zip_longest according to
    # https://pythonhosted.org/six/index.html?highlight=zip_longest but
    # that doesn't work for me.
    from itertools import izip_longest as zip_longest

# A list of the ambiguous values is given at
# https://en.wikipedia.org/wiki/Nucleic_acid_notation
AMBIGUOUS = {
    'A': {'A'},
    'C': {'C'},
    'G': {'G'},
    'T': {'T'},
    'M': {'A', 'C'},
    'R': {'A', 'G'},
    'W': {'A', 'T'},
    'S': {'G', 'C'},
    'K': {'G', 'T'},
    'Y': {'C', 'T'},
    'V': {'A', 'C', 'G'},
    'H': {'A', 'C', 'T'},
    'D': {'A', 'G', 'T'},
    'B': {'C', 'G', 'T'},
    'N': {'A', 'C', 'G', 'T'},
}

# Make a reverse version of AMBIGUOUS.
BASES_TO_AMBIGUOUS = dict(
    (''.join(sorted(bases)), symbol) for symbol, bases in AMBIGUOUS.items())


def matchToString(dnaMatch, read1, read2, matchAmbiguous=True, indent='',
                  offsets=None, includeGapLocations=True):
    """
    Format a DNA match as a string.

    @param dnaMatch: A C{dict} returned by C{compareDNAReads}.
    @param read1: A C{Read} instance or an instance of one of its subclasses.
    @param read2: A C{Read} instance or an instance of one of its subclasses.
    @param matchAmbiguous: If C{True}, ambiguous nucleotides that are
        possibly correct were counted as actually being correct. Otherwise,
        the match was done strictly, insisting that only non-ambiguous
        nucleotides could contribute to the matching nucleotide count.
    @param indent: A C{str} to indent all returned lines with.
    @param offsets: If not C{None}, a C{set} of offsets of interest that were
        only considered when making C{match}.
    @param includeGapLocations: If C{True} indicate the (1-based) locations of
        gaps.
    @return: A C{str} describing the match.
    """
    match = dnaMatch['match']
    identicalMatchCount = match['identicalMatchCount']
    ambiguousMatchCount = match['ambiguousMatchCount']
    gapMismatchCount = match['gapMismatchCount']
    gapGapMismatchCount = match['gapGapMismatchCount']
    nonGapMismatchCount = match['nonGapMismatchCount']

    if offsets:
        len1 = len2 = len(offsets)
    else:
        len1, len2 = map(len, (read1, read2))

    result = []
    append = result.append

    append(countPrint('%sExact matches' % indent, identicalMatchCount,
                      len1, len2))
    append(countPrint('%sAmbiguous matches' % indent, ambiguousMatchCount,
                      len1, len2))
    if ambiguousMatchCount and identicalMatchCount:
        anyMatchCount = identicalMatchCount + ambiguousMatchCount
        append(countPrint('%sExact or ambiguous matches' % indent,
                          anyMatchCount, len1, len2))

    mismatchCount = (gapMismatchCount + gapGapMismatchCount +
                     nonGapMismatchCount)
    append(countPrint('%sMismatches' % indent, mismatchCount, len1, len2))
    conflicts = 'conflicts' if matchAmbiguous else 'conflicts or ambiguities'
    append(countPrint('%s  Not involving gaps (i.e., %s)' % (indent,
                      conflicts), nonGapMismatchCount, len1, len2))
    append(countPrint('%s  Involving a gap in one sequence' % indent,
                      gapMismatchCount, len1, len2))
    append(countPrint('%s  Involving a gap in both sequences' % indent,
                      gapGapMismatchCount, len1, len2))

    for read, key in zip((read1, read2), ('read1', 'read2')):
        append('%s  Id: %s' % (indent, read.id))
        length = len(read)
        append('%s    Length: %d' % (indent, length))
        gapCount = len(dnaMatch[key]['gapOffsets'])
        append(countPrint('%s    Gaps' % indent, gapCount, length))
        if includeGapLocations and gapCount:
            append(
                '%s    Gap locations (1-based): %s' %
                (indent,
                 ', '.join(map(lambda offset: str(offset + 1),
                               sorted(dnaMatch[key]['gapOffsets'])))))
        ambiguousCount = len(dnaMatch[key]['ambiguousOffsets'])
        append(countPrint('%s    Ambiguous' % indent, ambiguousCount, length))
        extraCount = dnaMatch[key]['extraCount']
        if extraCount:
            append(countPrint('%s    Extra nucleotides at end' % indent,
                              extraCount, length))

    return '\n'.join(result)


def compareDNAReads(read1, read2, matchAmbiguous=True, gapChars='-',
                    offsets=None):
    """
    Compare two DNA sequences.

    @param read1: A C{Read} instance or an instance of one of its subclasses.
    @param read2: A C{Read} instance or an instance of one of its subclasses.
    @param matchAmbiguous: If C{True}, count ambiguous nucleotides that are
        possibly correct as actually being correct, and score these in the
        ambiguousMatchCount. Otherwise, we are strict and insist that only
        non-ambiguous nucleotides can contribute to the matching nucleotide
        count.
    @param gapChars: An object supporting __contains__ with characters that
        should be considered to be gaps.
    @param offsets: If not C{None}, a C{set} of offsets of interest. Offsets
        not in the set will not be considered.
    @return: A C{dict} with information about the match and the individual
        sequences (see below).
    """
    identicalMatchCount = ambiguousMatchCount = 0
    gapMismatchCount = nonGapMismatchCount = gapGapMismatchCount = 0
    read1ExtraCount = read2ExtraCount = 0
    read1GapOffsets = []
    read2GapOffsets = []
    read1AmbiguousOffsets = []
    read2AmbiguousOffsets = []
    empty = set()

    def _identicalMatch(a, b):
        return a == b and len(AMBIGUOUS[a]) == 1

    def _ambiguousMatch(a, b, matchAmbiguous):
        """
        Checks if two characters match ambiguously if matchAmbiguous is True.
        A match is an ambiguous match if it is not an identical match, but the
        sets of ambiguous characters overlap.
        """
        return (matchAmbiguous and
                not _identicalMatch(a, b) and
                AMBIGUOUS.get(a, empty) & AMBIGUOUS.get(b, empty))

    for offset, (a, b) in enumerate(zip_longest(read1.sequence.upper(),
                                                read2.sequence.upper())):
        # Use 'is not None' in the following to allow an empty offsets set
        # to be passed.
        if offsets is not None and offset not in offsets:
            continue
        if len(AMBIGUOUS.get(a, '')) > 1:
            read1AmbiguousOffsets.append(offset)
        if len(AMBIGUOUS.get(b, '')) > 1:
            read2AmbiguousOffsets.append(offset)
        if a is None:
            # b has an extra character at its end (it cannot be None).
            assert b is not None
            read2ExtraCount += 1
            if b in gapChars:
                read2GapOffsets.append(offset)
        elif b is None:
            # a has an extra character at its end.
            read1ExtraCount += 1
            if a in gapChars:
                read1GapOffsets.append(offset)
        else:
            # We have a character from both sequences (they could still be
            # gap characters).
            if a in gapChars:
                read1GapOffsets.append(offset)
                if b in gapChars:
                    # Both are gaps. This can happen (though hopefully not
                    # if the sequences were pairwise aligned).
                    gapGapMismatchCount += 1
                    read2GapOffsets.append(offset)
                else:
                    # a is a gap, b is not.
                    gapMismatchCount += 1
            else:
                if b in gapChars:
                    # b is a gap, a is not.
                    gapMismatchCount += 1
                    read2GapOffsets.append(offset)
                else:
                    # Neither is a gap character.
                    if _identicalMatch(a, b):
                        identicalMatchCount += 1
                    elif _ambiguousMatch(a, b, matchAmbiguous):
                        ambiguousMatchCount += 1
                    else:
                        nonGapMismatchCount += 1

    return {
        'match': {
            'identicalMatchCount': identicalMatchCount,
            'ambiguousMatchCount': ambiguousMatchCount,
            'gapMismatchCount': gapMismatchCount,
            'gapGapMismatchCount': gapGapMismatchCount,
            'nonGapMismatchCount': nonGapMismatchCount,
        },
        'read1': {
            'ambiguousOffsets': read1AmbiguousOffsets,
            'extraCount': read1ExtraCount,
            'gapOffsets': read1GapOffsets,
        },
        'read2': {
            'ambiguousOffsets': read2AmbiguousOffsets,
            'extraCount': read2ExtraCount,
            'gapOffsets': read2GapOffsets,
        },
    }


def findKozakConsensus(read):
    """
    In a given DNA sequence, search for a Kozak consensus: (gcc)gccRccATGG.
    The upper case bases in that pattern are required, and the lower case
    bases are the ones most frequently found at the given positions. The
    initial 'gcc' sequence (in parentheses) is of uncertain significance
    and is not taken into account here.

    @param read: A C{DNARead} instance to be checked for Kozak consensi.
    @return: A generator that yields C{DNAKozakRead} instances.
    """
    readLen = len(read)
    if readLen > 9:
        offset = 6
        readSeq = read.sequence
        while offset < readLen - 3:
            triplet = readSeq[offset:offset + 3]
            if triplet == 'ATG':
                if readSeq[offset + 3] == 'G':
                    if readSeq[offset - 3] in 'GA':
                        kozakQualityCount = sum((
                            readSeq[offset - 1] == 'C',
                            readSeq[offset - 2] == 'C',
                            readSeq[offset - 4] == 'C',
                            readSeq[offset - 5] == 'C',
                            readSeq[offset - 6] == 'G'))

                        kozakQualityPercent = kozakQualityCount / 5.0 * 100
                        yield DNAKozakRead(read, offset - 6, offset + 4,
                                           kozakQualityPercent)
            offset += 1



import six
from unittest import TestCase

from dark.colors import ColorsForCounts


class TestColorsForCounts(TestCase):
    """
    Test the ColorsForCounts class.
    """
    def testCountNotInt(self):
        """
        A C{ValueError} must be raised if a non-integer count is given.
        """
        error = (r"^color arguments must be given as space-separated pairs of "
                 r"\"count color\" where the count is an integer threshold\. "
                 r"Your value \('non-int'\) was not an integer\.$")
        six.assertRaisesRegex(self, ValueError, error, ColorsForCounts,
                              ('non-int red',))

    def testCountNegative(self):
        """
        A C{ValueError} must be raised if a count is less than zero.
        """
        error = (r"^color arguments must be given as space-separated pairs "
                 r"of \"count color\" where the count is non-negative. Your "
                 r"value \(-4\) is less than 0\.$")
        six.assertRaisesRegex(self, ValueError, error, ColorsForCounts,
                              ('-4 red',))

    def testNoSpace(self):
        """
        A C{ValueError} must be raised if a count/color value has no space.
        """
        error = (r"^color arguments must be given as space-separated pairs "
                 r"of \"value color\"\. Your value \('4red'\) does not "
                 r"contain a space\.$")
        six.assertRaisesRegex(self, ValueError, error, ColorsForCounts,
                              ('4red',))

    def testRepeatedCount(self):
        """
        A C{ValueError} must be raised if a count is repeated.
        """
        error = r"^repeated color argument count \(4\)\.$"
        six.assertRaisesRegex(self, ValueError, error, ColorsForCounts,
                              ('4 red', '4 black'))

    def testThresholdForCountNegative(self):
        """
        The thresholdForCount method must raise AssertionError if the count
        is negative.
        """
        colors = ColorsForCounts(('4 red',))
        error = (r"^Count \(-4\) cannot be negative\.$")
        six.assertRaisesRegex(self, AssertionError, error,
                              colors.thresholdForCount, -4)

    def testThresholdForTooLowCount(self):
        """
        The thresholdForCount method must work correctly when the lowest
        threshold is not reached.
        """
        colors = ColorsForCounts(('4 red',))
        self.assertEqual(0, colors.thresholdForCount(2))

    def testThresholdForCountOneThreshold(self):
        """
        The thresholdForCount method must work correctly when there is one
        threshold.
        """
        colors = ColorsForCounts(('4 red',))
        self.assertEqual(4, colors.thresholdForCount(4))
        self.assertEqual(4, colors.thresholdForCount(10))

    def testThresholdForCountTwoThresholds(self):
        """
        The thresholdForCount method must work correctly when there are two
        thresholds.
        """
        colors = ColorsForCounts(('4 red', '10 green'))
        self.assertEqual(4, colors.thresholdForCount(6))
        self.assertEqual(10, colors.thresholdForCount(12))

    def testThresholdForCountThreeThresholds(self):
        """
        The thresholdForCount method must work correctly when there are three
        thresholds.
        """
        colors = ColorsForCounts(('4 red', '10 green', '100 blue'))
        self.assertEqual(4, colors.thresholdForCount(6))
        self.assertEqual(10, colors.thresholdForCount(86))
        self.assertEqual(100, colors.thresholdForCount(1000))

    def testColorForTooLowCount(self):
        """
        The colorForCount method must work correctly when the lowest
        threshold is not reached.
        """
        colors = ColorsForCounts(('4 red',))
        self.assertEqual('black', colors.colorForCount(2))

    def testColorForTooLowCountDefault(self):
        """
        The colorForCount method must work correctly when the lowest
        threshold is not reached and a default color is given.
        """
        colors = ColorsForCounts(('4 red',), defaultColor='fuschia')
        self.assertEqual('fuschia', colors.colorForCount(2))

    def testColorForCountOneThreshold(self):
        """
        The colorForCount method must work correctly when there is one
        threshold.
        """
        colors = ColorsForCounts(('4 red',))
        self.assertEqual('black', colors.colorForCount(1))
        self.assertEqual('red', colors.colorForCount(4))
        self.assertEqual('red', colors.colorForCount(10))

    def testColorForCountTwoThresholds(self):
        """
        The colorForCount method must work correctly when there are two
        thresholds.
        """
        colors = ColorsForCounts(('4 red', '10 green'))
        self.assertEqual('black', colors.colorForCount(1))
        self.assertEqual('red', colors.colorForCount(6))
        self.assertEqual('green', colors.colorForCount(12))

    def testColorForCountThreeThresholds(self):
        """
        The colorForCount method must work correctly when there are three
        thresholds.
        """
        colors = ColorsForCounts(('4 red', '10 green', '100 blue'))
        self.assertEqual('black', colors.colorForCount(1))
        self.assertEqual('red', colors.colorForCount(6))
        self.assertEqual('green', colors.colorForCount(86))
        self.assertEqual('blue', colors.colorForCount(1000))

    def testThresholdToCssName(self):
        """
        The thresholdToCssName method must work as expected.
        """
        colors = ColorsForCounts(('4 red',))
        self.assertEqual('threshold-4', colors.thresholdToCssName(4))

    def testColorsForThreeThresholds(self):
        """
        The colors attribute must be as expected when there are three
        thresholds.
        """
        colors = ColorsForCounts(('4 red', '10 green', '100 blue'),
                                 defaultColor='pink')
        self.assertEqual(
            ((100, 'blue'), (10, 'green'), (4, 'red'), (0, 'pink')),
            colors.colors)



#!/usr/bin/env python

from __future__ import print_function, division

import argparse

from dark.proteins import SqliteIndex


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Query an sqlite3 protein/genome database and write '
                 'the matching sequences to standard output as FASTA.'))

parser.add_argument(
    '--databaseFile', required=True,
    help=('The protein/genome database file. This must be an sqlite3 database '
          'file created by make-protein-database.py'))

parser.add_argument(
    '--query',
    help=('An SQL database query. '
          'The query must result in rows with two columns, the sequence id '
          'and the sequence. These can be protein or nucleotide sequences, '
          'but it is probably better not to mix them!'))

parser.add_argument(
    '--queryFile',
    help=('The name of a file containing an SQL database query. '
          'The query must result in rows with two columns, the sequence id '
          'and the sequence. These can be protein or nucleotide sequences, '
          'but it is probably better not to mix them!'))

parser.add_argument(
    '--keepDuplicateIds', default=False, action='store_true',
    help='If specified, do not remove duplicate ids.')

parser.add_argument(
    '--idsOnly', default=False, action='store_true',
    help='If specified, only print matching sequence ids.')

args = parser.parse_args()

db = SqliteIndex(args.databaseFile)

if args.queryFile:
    query = open(args.queryFile).read()
elif args.query:
    query = args.query
else:
    raise RuntimeError('You must use either --query or --queryFile')

cur = db.execute(query)

removeDuplicates = not args.keepDuplicateIds
idsOnly = args.idsOnly

if removeDuplicates:
    seen = set()

while True:
    rows = cur.fetchmany()
    if rows:
        for row in rows:
            try:
                id_, sequence, genomeName = row
            except ValueError:
                id_, sequence = row
            else:
                id_ = '%s [%s]' % (id_, genomeName)

            if removeDuplicates:
                if id_ in seen:
                    continue
                else:
                    seen.add(id_)
            if idsOnly:
                print(id_)
            else:
                print('>%s\n%s' % (id_, sequence))
    else:
        break



#!/usr/bin/env python

from __future__ import print_function

import sys
import argparse
from itertools import chain

from dark.fasta import FastaReads
from dark.fastq import FastqReads


if __name__ == '__main__':

    # We do not use the addFASTACommandLineOptions and
    # parseFASTACommandLineOptions utility functions below because we allow
    # multiple FASTA or FASTQ files on the command line, which we specify
    # by --fasta and --fastq. And those names clash with the option names
    # used by those utility functions.

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=('Filter reads alignments to select a subset of reads. ',
                     'Given BLAST or DIAMOND JSON output files and the '
                     'FASTA (or FASTQ) query sequence files given to BLAST or '
                     'DIAMOND, and some filtering criteria, print the input '
                     'reads that pass the filtering.'))

    parser.add_argument(
        '--matcher', default='blast', choices=('blast', 'diamond'),
        help='The matching algorithm that was used to produce the JSON.')

    parser.add_argument(
        '--json', metavar='FILE.JSON[.BZ2]', nargs='+', action='append',
        required=True, help='the JSON file(s) of BLAST or DIAMOND output.')

    # A mutually exclusive group for either FASTA or FASTQ files.
    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        '--fasta', metavar='FILE.FASTA', nargs='+', action='append',
        help=('the FASTA file(s) of sequences that were given to BLAST '
              'or DIAMOND.'))

    group.add_argument(
        '--fastq', metavar='FILE.FASTQ', nargs='+', action='append',
        help=('the FASTQ file(s) of sequences that were given to BLAST '
              'or DIAMOND.'))

    # Args specific to DIAMOND.

    # A group for either the DIAMOND FASTA file or a sqlite3 database
    # of the FASTA used to make the DIAMOND database.
    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        '--diamondDatabaseFastaFilename', metavar='FILE.FASTA',
        help=('The filename of the FASTA file used to make the DIAMOND '
              'database. If --matcher diamond is used, either this argument '
              'or --diamondSqliteDatabaseFilename must be specified.'))

    group.add_argument(
        '--diamondSqliteDatabaseFilename', metavar='FILE.SQL',
        help=('The filename of the sqlite3 database file of FASTA metadata, '
              'made from the FASTA that was used to make the DIAMOND '
              'database. If --matcher diamond is used, either this argument '
              'or --diamondDatabaseFilename must be specified.'))

    parser.add_argument(
        '--diamondDatabaseFastaDirectory', metavar='DIR',
        help=('The directory where the FASTA file used to make the DIAMOND '
              'database can be found. This argument is only useful when '
              '--diamondSqliteDatabaseFilename is specified.'))

    # Args for filtering on ReadsAlignments.
    parser.add_argument(
        '--minStart', type=int, metavar='OFFSET',
        help='Reads that start before this subject offset should not be '
        'shown.')

    parser.add_argument(
        '--maxStop', type=int, metavar='OFFSET',
        help='Reads that end after this subject offset should not be shown.')

    parser.add_argument(
        '--oneAlignmentPerRead', default=False, action='store_true',
        help='If True, only keep the best alignment for each read.')

    parser.add_argument(
        '--maxAlignmentsPerRead', type=int, metavar='N',
        help=('Reads with more than this many alignments will be elided. Pass '
              'zero to only keep reads with no matches (alignments).'))

    parser.add_argument(
        '--scoreCutoff', type=float, metavar='SCORE',
        help=('A float score. Matches with scores worse than this will be '
              'ignored.'))

    parser.add_argument(
        '--maxHspsPerHit', type=int, metavar='N',
        help=('A numeric maximum number of HSPs to keep for each subject '
              'match.'))

    parser.add_argument(
        '--whitelist', nargs='+', action='append',
        metavar='TITLE', help='subject titles that should be whitelisted')

    parser.add_argument(
        '--blacklist', nargs='+', action='append',
        metavar='TITLE', help='subject titles that should be blacklisted')

    parser.add_argument(
        '--titleRegex', metavar='TITLE-REGEX',
        help='a regex that subject titles must match.')

    parser.add_argument(
        '--negativeTitleRegex', metavar='TITLE-REGEX',
        help='a regex that subject titles must not match.')

    parser.add_argument(
        '--truncateTitlesAfter',
        help=('a string that subject titles will be truncated beyond. If the '
              'truncated version of a title has already been seen, '
              'that title will be skipped.'))

    parser.add_argument(
        '--minSequenceLen', type=int, metavar='N',
        help='subjects of lesser length will be elided.')

    parser.add_argument(
        '--maxSequenceLen', type=int, metavar='N',
        help='subjects of greater length will be elided.')

    parser.add_argument(
        '--taxonomy', metavar='NAME',
        help=('the taxonomic group that subjects must match '
              'E.g., "Vira" will filter on viruses.'))

    args = parser.parse_args()

    # Flatten lists of lists that we get from using both nargs='+' and
    # action='append'. We use both because it allows people to use (e.g.)
    # --json on the command line either via "--json file1 --json file2" or
    # "--json file1 file2", or a combination of these. That way it's not
    # necessary to remember which way you're supposed to use it and you also
    # can't be hit by the subtle problem encountered in
    # https://github.com/acorg/dark-matter/issues/453
    jsonFiles = list(chain.from_iterable(args.json))
    whitelist = (
        set(chain.from_iterable(args.whitelist)) if args.whitelist else None)
    blacklist = (
        set(chain.from_iterable(args.blacklist)) if args.blacklist else None)

    # TODO: Add a --readClass command-line option in case we want to
    # process FASTA containing AA sequences.
    if args.fasta:
        reads = FastaReads(list(chain.from_iterable(args.fasta)))
    else:
        reads = FastqReads(list(chain.from_iterable(args.fastq)))

    if args.matcher == 'blast':
        from dark.blast.alignments import BlastReadsAlignments
        readsAlignments = BlastReadsAlignments(reads, jsonFiles)
    else:
        # Must be 'diamond' (due to parser.add_argument 'choices' argument).
        if (args.diamondDatabaseFastaFilename is None and
                args.diamondSqliteDatabaseFilename is None):
            print('Either --diamondDatabaseFastaFilename or '
                  '--diamondSqliteDatabaseFilename must be used with '
                  '--matcher diamond.', file=sys.stderr)
            sys.exit(1)
        elif not (args.diamondDatabaseFastaFilename is None or
                  args.diamondSqliteDatabaseFilename is None):
            print('--diamondDatabaseFastaFilename and '
                  '--diamondSqliteDatabaseFilename cannot both be used with '
                  '--matcher diamond.', file=sys.stderr)
            sys.exit(1)

        from dark.diamond.alignments import DiamondReadsAlignments
        readsAlignments = DiamondReadsAlignments(
            reads, jsonFiles,
            databaseFilename=args.diamondDatabaseFastaFilename,
            databaseDirectory=args.diamondDatabaseFastaDirectory,
            sqliteDatabaseFilename=args.diamondSqliteDatabaseFilename)

    readsAlignments.filter(
        maxAlignmentsPerRead=args.maxAlignmentsPerRead,
        minSequenceLen=args.minSequenceLen,
        maxSequenceLen=args.maxSequenceLen,
        minStart=args.minStart, maxStop=args.maxStop,
        oneAlignmentPerRead=args.oneAlignmentPerRead,
        maxHspsPerHit=args.maxHspsPerHit,
        scoreCutoff=args.scoreCutoff,
        whitelist=whitelist, blacklist=blacklist,
        titleRegex=args.titleRegex, negativeTitleRegex=args.negativeTitleRegex,
        truncateTitlesAfter=args.truncateTitlesAfter, taxonomy=args.taxonomy)

    format_ = 'fasta' if args.fasta else 'fastq'
    write = sys.stdout.write

    for readAlignments in readsAlignments:
        write(readAlignments.read.toString(format_=format_))



class DatabaseDuplicationError(Exception):
    """An attempt was made to insert a duplicate into a database."""


class NoSuchGenomeError(Exception):
    """A genome accession number could not be found."""


class NoSuchProteinError(Exception):
    """A protein accession number could not be found."""



"""
Probably the most comprehensive online reference for the way BLAST
calculates scores is "BLAST Scoring Parameters" by E. Michael Gertz

You can get a copy at
    ftp://ftp.ncbi.nlm.nih.gov/blast/documents/developer/scoring.pdf

or
    https://notebooks.antigenic-cartography.org/terry/docs/papers/\
    blast-scoring-parameters.pdf

Section 5 has the calculation of effective search space.

For details on going between bit scores and e-values, see
The Statistics of Sequence Similarity Scores, at
http://www.ncbi.nlm.nih.gov/BLAST/tutorial/Altschul-1.html
"""

from math import log

_LOG2 = log(2.0)


def bitScoreToEValue(bitScore, dbSize, dbSequenceCount, queryLength,
                     lengthAdjustment):
    """
    Convert a bit score to an e-value.

    @param bitScore: The C{float} bit score to convert.
    @param dbSize: The C{int} total size of the database (i.e., the sum of
        the lengths of all sequences in the BLAST database).
    @param dbSequenceCount: The C{int} number of sequences in the database.
    @param queryLength: The C{int} length of the query.
    @param lengthAdjustment: The C{int} length adjustment (BLAST XML output
        calls this the Statistics_hsp-len).
    @return: A C{float} e-value.
    """
    effectiveDbSize = (
        (dbSize - dbSequenceCount * lengthAdjustment) *
        (queryLength - lengthAdjustment)
    )
    return effectiveDbSize * (2.0 ** (-1.0 * bitScore))


def eValueToBitScore(eValue, dbSize, dbSequenceCount, queryLength,
                     lengthAdjustment):
    """
    Convert an e-value to a bit score.

    @param eValue: The C{float} e-value to convert.
    @param dbSize: The C{int} total size of the database (i.e., the sum of
        the lengths of all sequences in the BLAST database).
    @param dbSequenceCount: The C{int} number of sequences in the database.
    @param queryLength: The C{int} length of the query.
    @param lengthAdjustment: The C{int} length adjustment (BLAST XML output
        calls this the Statistics_hsp-len).
    @return: A C{float} bit score.
    """
    effectiveDbSize = (
        (dbSize - dbSequenceCount * lengthAdjustment) *
        (queryLength - lengthAdjustment)
    )
    return -1.0 * (log(eValue / effectiveDbSize) / _LOG2)



#!/usr/bin/env python

from __future__ import print_function, division

import argparse
import requests

# See https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EFetch
# for URL format details.

URL = ('https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?'
       'db=%(database)s&id=%(id)s&rettype=fasta&retmode=text')

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Fetch a sequence by id from NCBI and write it to stdout '
                 'as FASTA.'))

parser.add_argument(
    'id', help='The id of the sequence to fetch.')

parser.add_argument(
    '--database', default='nucleotide', choices=('nucleotide', 'protein'),
    help='The name of the NCBI database to query.')

args = parser.parse_args()

print(requests.get(
    URL % {'database': args.database, 'id': args.id}
).text.rstrip('\n'))



#!/usr/bin/env python

from __future__ import print_function

import sys
import argparse
from os.path import exists, join
from os import mkdir
from tempfile import mkdtemp

from dark.process import Executor


def makeOuputDir(outputDir, force):
    """
    Create or check for an output directory.

    @param outputDir: A C{str} output directory name, or C{None}.
    @param force: If C{True}, allow overwriting of pre-existing files.
    @return: The C{str} output directory name.
    """
    if outputDir:
        if exists(outputDir):
            if not force:
                print('Will not overwrite pre-existing files. Use --force to '
                      'make me.', file=sys.stderr)
                sys.exit(1)
        else:
            mkdir(outputDir)
    else:
        outputDir = mkdtemp()
        print('Writing output files to %s' % outputDir)

    return outputDir


def samtoolsMpileup(outFile, referenceFile, alignmentFile, executor):
    """
    Use samtools mpileup to generate VCF.

    @param outFile: The C{str} name to write the output to.
    @param referenceFile: The C{str} name of the FASTA file with the reference
        sequence.
    @param alignmentFile: The C{str} name of the SAM or BAM alignment file.
    @param executor: An C{Executor} instance.
    """
    executor.execute(
        'samtools mpileup -u -v -f %s %s > %s' %
        (referenceFile, alignmentFile, outFile))


def bcftoolsMpileup(outFile, referenceFile, alignmentFile, executor):
    """
    Use bcftools mpileup to generate VCF.

    @param outFile: The C{str} name to write the output to.
    @param referenceFile: The C{str} name of the FASTA file with the reference
        sequence.
    @param alignmentFile: The C{str} name of the SAM or BAM alignment file.
    @param executor: An C{Executor} instance.
    """
    executor.execute(
        'bcftools mpileup -Ov -f %s %s > %s' %
        (referenceFile, alignmentFile, outFile))


def bcftoolsCallMulti(outFile, vcfFile, executor):
    """
    Use bcftools to make consensus calls, using the  -m (multiallelic) option.

    @param outFile: The C{str} name to write the output to.
    @param vcfFile: The C{str} name of the VCF file with the calls from
        the pileup.
    @param executor: An C{Executor} instance.
    """
    # Note that this just gives a call on the first genome in the calls file.
    executor.execute(
        'bcftools call -m -Ov -o %s < %s' % (outFile, vcfFile))


def bcftoolsCallConsensus(outFile, vcfFile, executor):
    """
    Use bcftools to make consensus calls, using the original -c option.

    @param outFile: The C{str} name to write the output to.
    @param vcfFile: The C{str} name of the VCF file with the calls from
        the pileup.
    @param executor: An C{Executor} instance.
    """
    # Note that this just gives a call on the first genome in the calls file.
    executor.execute(
        'bcftools call -c -Ov -o %s < %s' % (outFile, vcfFile))


def bcftoolsConsensus(outFile, vcfFile, id_, referenceFile, executor):
    """
    Use bcftools to extract consensus FASTA.

    @param outFile: The C{str} name to write the output to.
    @param vcfFile: The C{str} name of the VCF file with the calls from
        the pileup.
    @param id_: The C{str} identifier to use in the resulting FASTA sequence.
    @param referenceFile: The C{str} name of the FASTA file with the reference
        sequence.
    @param executor: An C{Executor} instance.
    """
    bgz = vcfFile + '.gz'
    executor.execute('bgzip -c %s > %s' % (vcfFile, bgz))
    executor.execute('tabix %s' % bgz)
    executor.execute(
        'bcftools consensus %s < %s | '
        'filter-fasta.py --idLambda \'lambda id: "%s"\' > %s' %
        (bgz, referenceFile, id_, outFile))


def vcfutilsConsensus(outFile, vcfFile, id_, _, executor):
    """
    Use vcftools to extract consensus FASTA.

    @param outFile: The C{str} name to write the output to.
    @param vcfFile: The C{str} name of the VCF file with the calls from
        the pileup.
    @param id_: The C{str} identifier to use in the resulting FASTA sequence.
    @param executor: An C{Executor} instance.
    """
    executor.execute(
        'vcfutils.pl vcf2fq < %s | '
        'filter-fasta.py --fastq --quiet --saveAs fasta '
        '--idLambda \'lambda id: "%s"\' > %s' %
        (vcfFile, id_, outFile))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Compare different consensus making methods.')

    parser.add_argument(
        '--referenceFile', required=True, metavar='FILENAME',
        help='The name of the FASTA file containing the reference sequence.')

    parser.add_argument(
        '--alignmentFile', required=True, metavar='FILENAME',
        help=('The name of the SAM or BAM file containing an alignment to the '
              'reference.'))

    parser.add_argument(
        '--verbose', type=int, default=0, metavar='N',
        help=('The integer verbosity level (0 = no output, 1 = some output, '
              '2 = maximal output).'))

    parser.add_argument(
        '--force', default=False, action='store_true',
        help='If given, overwrite pre-existing files.')

    parser.add_argument(
        '--outputDir', metavar='DIRNAME',
        help='The directory to save result files to.')

    args = parser.parse_args()

    outputDir = makeOuputDir(args.outputDir, args.force)

    executor = Executor()

    pileuppers = (
        ('samtools-mpileup', samtoolsMpileup),
        ('bcftools-mpileup', bcftoolsMpileup))

    callers = (
        ('bcftools-c', bcftoolsCallConsensus),
        ('bcftools-m', bcftoolsCallMulti))

    consensusers = (
        ('bcftools-consensus', bcftoolsConsensus),
        ('vcfutils-vcf2fq', vcfutilsConsensus))

    consensusFiles = []

    for pileupName, pileupFunc in pileuppers:
        pileupMiddle = pileupName
        pileupFile = join(outputDir, 'pileup-' + pileupMiddle + '.vcf')
        pileupFunc(pileupFile, args.referenceFile,
                   args.alignmentFile, executor)

        for callerName, callerFunc in callers:
            callerMiddle = pileupMiddle + '-' + callerName
            callerFile = join(outputDir, 'calls-' + callerMiddle + '.vcf')
            callerFunc(callerFile, pileupFile, executor)

            for consensusName, consensusFunc in consensusers:
                consensusMiddle = callerMiddle + '-' + consensusName
                consensusFile = join(outputDir,
                                     'consensus-' + consensusMiddle + '.fasta')
                consensusFiles.append(consensusFile)
                consensusFunc(consensusFile, callerFile, consensusMiddle,
                              args.referenceFile, executor)

    # Let's assume there's at least one consensus file.
    consensusesFile = join(outputDir, 'consensuses.fasta')
    executor.execute(
        'cat %s > %s' % (' '.join(consensusFiles), consensusesFile))

    htmlFile = join(outputDir, 'consensus-identity.html')
    executor.execute(
        ('fasta-identity-table.py --footer --showGaps --showLengths < %s | '
         "perl -pe 's/-(bcftools|vcfutils)/ $1/g' > %s") %
        (consensusesFile, htmlFile))

    verbose = args.verbose
    if verbose > 0:
        print('The following commands were executed:')
        for line in executor.log:
            if line.startswith('#'):
                if verbose > 1:
                    print(line)
            else:
                print(line)

    print('Identity table comparing %d consensuses written to %s' %
          (len(consensusFiles), htmlFile))



from collections import Counter

from dark.errors import NoSuchGenomeError
from dark.genbank import GenomeRanges
from dark.reads import DNARead
from dark.sam import samfile


class GenomeProteinInfo(object):
    """
    Hold information about the proteins in a genome and how they are matched
    by reads in SAM files.

    @param accession: The C{str} accession number of a genome.
    @param proteinGenomeDB: A L{dark.civ.proteins.SqliteIndex} instance.
    @param checkTranslations: If C{True}, check that the protein sequences
        that area supposed to have come from the genome can be obtained by
        translating the corresponding region of the genome.
    """
    def __init__(self, genomeAccession, proteinGenomeDB,
                 checkTranslations=True):
        self.genomeAccession = genomeAccession
        self.proteinGenomeDB = proteinGenomeDB
        # self.proteins is keyed by protein accession number.
        self.proteins = {}
        self.coveredProteins = set()
        # self.offsets is keyed by genome offset, values are dicts that
        # contain a list of protein accession numbers that overlap that
        # offset and a set of read ids (if any) that match at that offset.
        # The offsets keys are only those that correspond to one or more
        # proteins in the genome.
        self.offsets = {}
        # self.coveredOffsetCount holds the read counts for all offsets covered
        # by reads, regardless of whether the offsets correspond to proteins or
        # not.
        self.coveredOffsetCount = Counter()
        self.samFiles = []
        self.readIdsMatchingGenome = set()

        self.genome = proteinGenomeDB.findGenome(genomeAccession)
        if self.genome is None:
            raise NoSuchGenomeError('Reference %r not found in protein/genome '
                                    'database.' % genomeAccession)

        for protein in proteinGenomeDB.findProteinsForGenome(genomeAccession):
            proteinAccession = protein['accession']
            self.proteins[proteinAccession] = protein

            ranges = GenomeRanges(protein['offsets']).ranges
            # print('Protein accession', proteinAccession)
            # print(ranges)

            for (start, stop, forward) in ranges:
                for offset in range(start, stop):
                    if offset not in self.offsets:
                        self.offsets[offset] = {
                            'proteinAccessions': set(),
                            'readIds': set(),
                        }
                    self.offsets[offset]['proteinAccessions'].add(
                        proteinAccession)

            if checkTranslations:
                self._checkTranslation(self.genome, ranges, protein)

    def _checkTranslation(self, genome, ranges, protein):
        """
        Make sure all protein sequences supposed to be in the genome can in
        fact be obtained by translating the genome.

        @param genome: A C{dict} with genome information from our sqlite3
            protein/genome database, as returned by
            C{dark.civ.proteins.SqliteIndex.findGenome.
        @param ranges: A C{list} of (start, stop, forward) nucleotide ranges
            for the protein in the genome.
        @param protein: A C{dict} with protein information from our sqlite3
            protein/genome database, as returned by
            C{dark.civ.proteins.SqliteIndex.findProtein.
        """
        proteinSequence = protein['sequence'] + '*'

        # print('protein name', protein['product'], 'ranges', ranges)
        sequence = ''.join([genome['sequence'][start:stop]
                            for (start, stop, _) in ranges])

        genomeRead = DNARead('id', sequence)
        translations = list(genomeRead.translations())
        index = 0 if protein['forward'] else 3
        if translations[index].sequence != proteinSequence:
            # TODO: improve this error to show what actually went wrong.
            raise ValueError(
                'Could not translate genome range to get protein sequence')

    def addSAM(self, filename, filterAlignment=None):
        """
        Read a SAM file and add information about the reads that match our
        reference id.

        @param filename: A C{str} SAM filename.
        @param filterAlignment: A 1-argument function to be used for filtering
            reads in the SAM file. If C{None}, all alignments will be examined.
        """
        self.samFiles.append(filename)
        referenceId = self.genomeAccession
        with samfile(filename) as sam:
            for column in sam.pileup():
                for read in column.pileups:
                    alignment = read.alignment
                    if (alignment.reference_name == referenceId and
                            (filterAlignment is None or
                             filterAlignment(alignment))):
                        readId = alignment.query_name
                        self.readIdsMatchingGenome.add(readId)
                        offset = column.reference_pos
                        self.coveredOffsetCount[offset] += 1

                        try:
                            offsetInfo = self.offsets[offset]
                        except KeyError:
                            pass
                        else:
                            # This offset corresponds to one or more proteins.
                            self.coveredProteins.update(
                                offsetInfo['proteinAccessions'])
                            offsetInfo['readIds'].add(readId)

    def proteinCoverageInfo(self, proteinAccession, minReadOffsetCount=None):
        """
        Calculate coverage information for a protein.

        @param proteinAccession: A C{str} accession number.
        @param minReadOffsetCount: An C{int}, specifying the minimum number of
            reads offsets that must overlap the protein for the read to be
            considered as sufficiently intersecting the protein. Use this to
            prevent reads that just overlap the protein in a very small number
            offsets from being counted. Or C{None} to indicate that no such
            filtering should be applied.
        @raises KeyError: If C{proteinAccession} is not known.
        @return: A C{dict} containing
                * the number of covered offsets,
                * the total number of read bases that cover the protein,
                * the protein length (in nucleotides), and
                * the set of all matching read ids.
            See below for the dictionary keys.
        """
        protein = self.proteins[proteinAccession]
        coveredOffsets = 0
        totalBases = 0
        allReadIds = set()
        offsetsSeen = set()
        proteinLength = 0

        if minReadOffsetCount is not None and minReadOffsetCount < 2:
            # A minimum of zero or one is equivalent to not giving a value.
            minReadOffsetCount = None

        if minReadOffsetCount:
            readOffsetCounts = Counter()

        proteinRanges = GenomeRanges(protein['offsets']).ranges

        # Do an initial pass across all the offsets of the protein to see
        # which reads intersect and where. We will then do a second pass in
        # which we ignore reads that do not sufficiently overlap.
        for (start, stop, forward) in proteinRanges:
            proteinLength += stop - start
            for offset in range(start, stop):
                assert offset not in offsetsSeen
                offsetsSeen.add(offset)
                readIds = self.offsets[offset]['readIds']
                if readIds and minReadOffsetCount:
                    readOffsetCounts.update(readIds)

        # Sanity check that the sum of the range lengths is the same as the
        # overall length given in the database.
        #
        # The +3 in the following is because the database holds the AA
        # length, not including the stop codon. But the database range
        # covers the stop codon.
        dbProteinLength = self.proteins[proteinAccession]['length'] * 3 + 3
        if proteinLength != dbProteinLength:
            raise ValueError(
                'Sum of protein database ranges (%d) does not agree with '
                'database protein length (%d) for protein %s!' %
                (proteinLength, dbProteinLength, proteinAccession))

        # If we are not reporting reads whose overlapping offset count is
        # too low, make a set of such reads.
        if minReadOffsetCount:
            unwanted = set(readId for readId in readOffsetCounts
                           if readOffsetCounts[readId] < minReadOffsetCount)
        else:
            unwanted = set()

        # Second pass, in which we ignore unwanted (i.e., insufficiently
        # overlapping) reads.
        for (start, stop, forward) in proteinRanges:
            for offset in range(start, stop):
                readIds = set(self.offsets[offset]['readIds']) - unwanted
                if readIds:
                    allReadIds.update(readIds)
                    coveredOffsets += 1
                    totalBases += len(readIds)

        return {
            'coveredOffsets': coveredOffsets,
            'totalBases': totalBases,
            'ntLength': proteinLength,
            'readIds': allReadIds,
        }

    def readIdsForAllProteins(self):
        """
        Get the set of read ids for reads matching any protein at all.

        @return: A C{set} of C{str} read ids.
        """
        result = set()
        for offsetInfo in self.offsets.values():
            result.update(offsetInfo['readIds'])
        return result



from __future__ import print_function

from os.path import join

from dark.fastq import FastqReads
from dark.html import NCBISequenceLinkURL


class AlignmentPanelHTMLWriter(object):
    """
    Produces HTML details of a rectangular panel of graphs that each
    contain an alignment graph against a given sequence. This is
    supplementary output info for the AlignmentPanel class in graphics.py.

    @param outputDir: The C{str} directory to write files into.
    @param titlesAlignments: A L{dark.titles.TitlesAlignments} instance.
    """
    def __init__(self, outputDir, titlesAlignments):
        self._outputDir = outputDir
        self._titlesAlignments = titlesAlignments
        self._images = []

    def addImage(self, imageBasename, accession, title, graphInfo):
        self._images.append({
            'accession': accession,
            'graphInfo': graphInfo,
            'imageBasename': imageBasename,
            'title': title
        })

    def close(self):
        with open(join(self._outputDir, 'index.html'), 'w') as fp:
            self._writeHeader(fp)
            self._writeBody(fp)
            self._writeFooter(fp)
        with open(join(self._outputDir, 'style.css'), 'w') as fp:
            self._writeCSS(fp)

    def _writeHeader(self, fp):
        fp.write("""\
<html>
  <head>
    <title>Read alignments for %d matched subjects</title>
    <link rel="stylesheet" type="text/css" href="style.css">
  </head>
  <body>
    <div id="content">
        """ % len(self._images))

    def _writeBody(self, fp):
        fp.write('<h1>Read alignments for %d matched subjects</h1>\n' %
                 len(self._images))

        # Write out an alignment panel as a table.
        cols = 6
        fp.write('<table><tbody>\n')

        for i, image in enumerate(self._images):
            title = image['title']
            accession = image['accession']
            if i % cols == 0:
                fp.write('<tr>\n')

            fp.write(
                '<td><a id="small_%s"></a><a href="#big_%s"><img src="%s" '
                'class="thumbnail"/></a></td>\n' %
                (accession, accession, image['imageBasename']))

            if i % cols == cols - 1:
                fp.write('</tr>')

        # Add empty cells to the final table row, and close the row, if
        # necessary.
        if i % cols < cols - 1:
            while i % cols < cols - 1:
                fp.write('<td>&nbsp;</td>\n')
                i += 1
            fp.write('</tr>\n')

        fp.write('</tbody></table>\n')

        # Write out the full images with additional detail.
        for i, image in enumerate(self._images):
            title = image['title']
            accession = image['accession']
            titleAlignments = self._titlesAlignments[title]
            graphInfo = image['graphInfo']
            readFormat = self._writeReads(image)
            fp.write("""
      <a id="big_%s"></a>
      <h3>%d: %s</h3>
      <p>
            Length: %d.
            Read count: %d.
            HSP count: %d.
            <a href="%s.%s">%s</a>.
            <a href="#small_%s">Top panel.</a>
"""
                     % (accession,
                        i, title,
                        titleAlignments.subjectLength,
                        titleAlignments.readCount(),
                        titleAlignments.hspCount(),
                        accession, readFormat, readFormat,
                        accession))

            url = NCBISequenceLinkURL(title)
            if url:
                fp.write('<a href="%s" target="_blank">NCBI</a>.' % url)

            # Write out feature information.
            if graphInfo['features'] is None:
                # Feature lookup was False (or we were offline).
                pass
            elif len(graphInfo['features']) == 0:
                fp.write('There were no features.')
            else:
                fp.write('<a href="%s">Features</a>' %
                         self._writeFeatures(i, image))

            # Write out the titles that this title invalidated due to its
            # read set.
            readSetFilter = self._titlesAlignments.readSetFilter
            if readSetFilter:
                invalidated = readSetFilter.invalidates(title)
                if invalidated:
                    nInvalidated = len(invalidated)
                    fp.write('<br/>This title invalidated %d other%s due to '
                             'its read set:<ul>'
                             % (nInvalidated,
                                '' if nInvalidated == 1 else 's'))
                    for title in invalidated:
                        fp.write('<li>%s</li>' % title)
                    fp.write('</ul>')

            fp.write('</p><img src="%s" class="full-size"/>' %
                     image['imageBasename'])

    def _writeFooter(self, fp):
        fp.write("""\
    </div>
  </body>
</html>
""")

    def _writeCSS(self, fp):
        fp.write("""\
#content {
  width: 95%;
  margin: auto;
}
img.thumbnail {
  height: 300px;
}
img.full-size {
  height: 900px;
}
""")

    def _writeReads(self, image):
        """
        Write a FASTA or FASTQ file containing the set of reads that hit a
        sequence.

        @param image: A member of self._images.
        @return: A C{str}, either 'fasta' or 'fastq' indicating the format
            of the reads in C{self._titlesAlignments}.
        """
        if isinstance(self._titlesAlignments.readsAlignments.reads,
                      FastqReads):
            format_ = 'fastq'
        else:
            format_ = 'fasta'
        filename = join(self._outputDir,
                        '%s.%s' % (image['accession'], format_))
        titleAlignments = self._titlesAlignments[image['title']]
        with open(filename, 'w') as fp:
            for titleAlignment in titleAlignments:
                fp.write(titleAlignment.read.toString(format_))
        return format_

    def _writeFeatures(self, image):
        """
        Write a text file containing the features as a table.

        @param image: A member of self._images.
        @return: The C{str} features file name - just the base name, not
            including the path to the file.
        """
        basename = image['accession'] + '-features.txt'
        filename = join(self._outputDir, basename)
        featureList = image['graphInfo']['features']
        # Note that the following (deliberately) creates an empty features
        # file if there were no features.
        with open(filename, 'w') as fp:
            for feature in featureList:
                fp.write('%s\n\n' % feature.feature)
        return basename



#!/usr/bin/env python

from __future__ import print_function

import argparse
from collections import defaultdict, Counter

from dark.filter import (
    addFASTAFilteringCommandLineOptions, parseFASTAFilteringCommandLineOptions)
from dark.reads import Reads
from dark.sam import samfile, SAMFilter

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description='Print SAM/BAM file coverage statistics.')

addFASTAFilteringCommandLineOptions(parser)
SAMFilter.addFilteringOptions(parser, samfileIsPositional=True)

args = parser.parse_args()

# We don't have a file of reads, we just want a read filter that we
# can use to filter the SAM file query sequences.
reads = parseFASTAFilteringCommandLineOptions(args, Reads())
samFilter = SAMFilter.parseFilteringOptions(args, reads.filterRead)

coveredOffsets = defaultdict(Counter)
coveringReads = defaultdict(set)

with samfile(args.samfile) as sam:
    for column in sam.pileup():
        referenceOffset = column.reference_pos
        for read in column.pileups:
            if samFilter.filterAlignment(read.alignment):
                referenceId = read.alignment.reference_name
                coveredOffsets[referenceId][referenceOffset] += 1
                coveringReads[referenceId].add(read.alignment.query_name)

referenceLengths = samFilter.referenceLengths()

for referenceId in sorted(referenceLengths):
    offsetsCovered = len(coveredOffsets[referenceId])
    referenceLength = referenceLengths[referenceId]
    print('%s: length %d, covering reads %d, covered sites %d (%.4f%%), '
          'coverage depth %.4f' %
          (referenceId, referenceLength, len(coveringReads[referenceId]),
           offsetsCovered, offsetsCovered / referenceLength * 100.0,
           sum(coveredOffsets[referenceId].values()) / referenceLength))



#!/usr/bin/env python

import sys
import argparse
from tempfile import mkdtemp
from os.path import join

from dark.fasta import FastaReads
from dark.process import Executor


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Make a consensus sequence.')

    parser.add_argument(
        '--reference', required=True,
        help='The reference FASTA file.')

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        '--bam',
        help=('The BAM file from which the consensus should be made. Only '
              'used if a VCF file is not provided.'))

    group.add_argument(
        '--vcfFile',
        help=('The VCF file. If omitted, bcftools will be used to make a VCF '
              'file from the BAM file.'))

    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        '--id',
        help=('The id to use in the consensus sequence in the output FASTA. '
              'If not given, the reference sequence id will be used.'))

    group.add_argument(
        '--idLambda', metavar='LAMBDA-FUNCTION',
        help=('A one-argument function taking and returning a read id. '
              'This can be used to set the id of the reference sequence based '
              'on the id of the reference sequence (the function will be '
              'called with the id of the reference sequence). E.g., '
              '--idLambda "lambda id: id.split(\'_\')[0]" or '
              '--idLambda "lambda id: id[:10] + \'-consensus\'".'))

    parser.add_argument(
        '--sample',
        help=('The name of the sample (from the @RG SM tag in the original '
              'alignment BAM file) for which a consensus should be made. '
              'If not given, the first sample name (from the #CHROM header) '
              'in the VCF file will be used.'))

    parser.add_argument(
        '--dryRun', default=False, action='store_true',
        help='Do not run commands, just print what would be done.')

    parser.add_argument(
        '--log', default=False, action='store_true',
        help=('Show a log of commands that were (or would be, if --dryRun is '
              'used) executed.'))

    parser.add_argument(
        '--noClean', default=True, action='store_false', dest='clean',
        help=('Do not remove intermediate files or the temporary directory.'))

    args = parser.parse_args()

    if args.bam and args.vcfFile:
        print('Only of --bam or --vcfFile can be given.', file=sys.stderr)
        sys.exit(0)

    if not (args.bam or args.vcfFile):
        print('One of --bam or --vcfFile must be given.', file=sys.stderr)
        sys.exit(0)

    e = Executor(args.dryRun)

    tempdir = mkdtemp(prefix='consensus-')

    if args.bam:
        # No VCF file provided, so make one.
        vcfFile = join(tempdir, 'vcf.gz')
        e.execute("bcftools mpileup --max-depth 5000 -Ou -f '%s' '%s' | "
                  "bcftools call -mv -Oz -o '%s'" %
                  (args.reference, args.bam, vcfFile))

        e.execute("bcftools index '%s'" % vcfFile)
    else:
        vcfFile = args.vcfFile

    if args.sample:
        sample = args.sample
    else:
        result = e.execute(
            "gunzip -c '%s' | egrep -m 1 '^#CHROM' | cut -f10" % vcfFile)
        sample = result.stdout.strip()

    consensusFile = join(tempdir, 'consensus.fasta')
    result = e.execute(
        "bcftools consensus --sample '%s' --iupac-codes --fasta-ref "
        "'%s' '%s' > '%s'" %
        (sample, args.reference, vcfFile, consensusFile))

    consensus = list(FastaReads(consensusFile))[0]

    if args.id is not None:
        consensus.id = args.id
    elif args.idLambda is not None:
        idLambda = eval(args.idLambda)
        consensus.id = idLambda(consensus.id)

    print(consensus.toString('fasta'), end='')

    if result.stderr:
        print(result.stderr, end='', file=sys.stderr)

    if args.dryRun or args.log:
        print('\n'.join(e.log), file=sys.stderr)

    if tempdir:
        if args.clean:
            e.execute("rm -r '%s'" % tempdir)
        else:
            print('Temporary directory %r.' % tempdir, file=sys.stderr)


if __name__ == '__main__':
    main()
